# -*- yform.py: python ; coding: utf-8 -*-
#####################################################
# 表单式输入组件库
#####################################################
from PySide6.QtWidgets import QWidget, QLayout, QVBoxLayout, QGroupBox

from .wdg import (TabSetting, TabInfo, YStackedPages, YTitle)
from .yfield import FACTORY, YFieldWidget, YFdCriteria
from .ymodel import PandasModel
from yut import format_obj
from . import EditMode, Stage, create_button, create_toolbar
from . import create_layout


############################################################
# 单笔数据录入/显示表单
############################################################
class YForm(QWidget):
    def __init__(self,
                 model: PandasModel = None,
                 column_specs=None,
                 title='标题未命名',
                 parent=None,
                 caption=None,
                 read_only=False,
                 edit_mode: EditMode = EditMode.MODIFY,
                 page_def: TabSetting = None,
                 layout_flag='V',
                 ):
        super().__init__(parent=parent)
        self.column_specs = column_specs
        self.caption = caption
        self.edit_mode = edit_mode
        self.read_only = read_only
        self.field_editors = []
        self.model = model
        self._data = {}
        if self.model:
            self.column_specs = model.column_specs()
            self._data = self.model.data_of_current_row()

        if not page_def:
            page_def = TabSetting(TabInfo('详细内容', layout_flag='F'))

        self.title = format_obj(self._data, title) if title else ''

        self.pages_widget = YStackedPages(page_def, parent=self)

        self.label_title = self.create_label_title()
        self.label_caption = self.create_label_caption()
        self.setWindowTitle(self.title)
        self.setLayout(create_layout(layout_flag))
        self.setup_ui()

    def refresh_display(self):
        # set caption
        if self.caption:
            self.label_caption.set_text(format_obj(self._data, self.caption))

    def setup_ui(self):
        m_layout = self.layout()
        # m_layout.setContentsMargins(12, 20, 12, 20)
        # 标题
        m_layout.addWidget(self.label_title)

        # 顶部标题和工具栏
        m_layout.addLayout(self.create_top_bar())

        # 页签组件
        m_layout.addWidget(self.pages_widget)

        # 更新绑定数据
        self.refresh_display()

    def clean(self):
        self.field_editors.clear()
        self.model = None

    def create_top_bar(self):
        tbox = create_layout('H')
        tbox.setContentsMargins(0, 8, 0, 16)
        tbox.addWidget(self.label_caption)
        tbox.addStretch(0)
        if not self.read_only:
            tbox.addWidget(self.create_toolbar())
        return tbox

    def add_editors(self, editors, page_index=None, stretch=True):
        layout = self.pages_widget.page_layout(page_index)
        for fd in editors:
            fd.bind(self._data)
            fd.add_to(layout)
        if hasattr(layout, "addStretch") and stretch:
            layout.addStretch(0)

    def add_fields(self, field_names: list, page_index=None, stretch=True):
        if self.read_only:  # 只读显示，使用只读部件
            items = [YFieldWidget(parent=self, col_spec=self.model.column_spec(fd)) for fd in field_names]
        else:
            items = [FACTORY.get_editor(self, self.model.column_spec(fd),
                                        stage=Stage.IN_FORM,
                                        read_only=False, )
                     for fd in field_names]
        self.add_editors(items, page_index, stretch)
        self.field_editors.extend(items)

    def build_default_fields(self):
        """
        按照默认内容添加字段控件。不分页签，将全部默认field加入其中
        :return:
        """
        self.add_fields([cp.name for cp in self.column_specs])

    def update_data(self):
        self.update_data_to(self.model.current_row_number())

    def update_data_to(self, row):
        self.model.set_current_row_number(row)
        self._data = self.model.data_of_current_row()
        for item in self.field_editors:
            item.bind(self._data)
        self.refresh_display()

    def data(self):
        for item in self.field_editors:
            item.assign_to(self._data)
        return self._data

    def create_label_title(self):
        return YTitle(self.title, parent=self, font_size=12, font_bold=True, icon_path=None)

    def create_label_caption(self):
        return YTitle(self.caption, parent=self, font_size=10, font_bold=True)

    def create_toolbar(self):
        buttons = [
            create_button('刷新',
                          trigger=self.load_data,
                          flat=True,
                          icon_file='refresh.png'),
            create_button('保存',
                          trigger=self.save_data,
                          tip_text='保存当前数据',
                          flat=True,
                          icon_file='save.png'),
        ]
        if self.edit_mode == EditMode.MODIFY:  # 修改模式，增加删除按钮
            buttons.append(create_button('删除',
                                         trigger=self.remove_data,
                                         tip_text='删除当前数据',
                                         flat=True,
                                         icon_file='cancel.png'))
        tbar = create_toolbar(buttons)
        return tbar

    def load_data(self):
        print(f'<YForm.load_data>  ...')

    def save_data(self):
        print(f'<YForm.save_data> saving ui data  ...', self._data)

    def remove_data(self):
        print(f'<YForm.remove_data> removing data  ...')


################################################################
# 查询条件表单
################################################################
class YQueryForm(QGroupBox):
    def __init__(self, parent=None, layout_flag='V', page_def: TabSetting = None, auto_from=None):
        super().__init__(parent)
        # 不要显示groupbox的顶部标题
        self.setObjectName('_query_form')
        self.setTitle("")
        self.setStyleSheet("#_query_form { margin-top: 0px; }")  # 减少顶部边距
        self.layout_flag = layout_flag
        if not page_def:
            page_def = TabSetting(TabInfo('查询条件', layout_flag=self.layout_flag))
        self.widget = YStackedPages(page_def, parent=self)
        self.items = []
        self.auto_from = auto_from
        self.setLayout(QVBoxLayout())
        self.layout().addWidget(self.widget)

    def add_criteria(self, items: list[YFdCriteria], page_index=None):
        layout: QLayout = self.widget.page_layout(page_index)
        for item in items:
            if self.auto_from:
                item.assign_from(self.auto_from)
            item.add_to(layout)
            self.items.append(item)
        if hasattr(layout, 'addStretch'):
            layout.addStretch(0)

    def assign_from(self, value):
        for item in self.items:
            item.assign_from(value)

    def assign_to(self, value):
        for item in self.items:
            item.assign_to(value)
        return value

    def values(self):
        return self.assign_to(dict())



# -*- ndw_models.py: python ; coding: utf-8 -*-
# ***********************************************************************
# 创建数据model对应的单表Selection对象
# 本代码由程序通过反射数据库表结构自动生成，模板文件： db_models.ptl (V1.0)
# 生成时间: 2025-06-12 23:26:52.988021
# ***********************************************************************
from yus.dbm import QueryManager, MetaDb, Selection

_ndw_db = None
NAME_SPACE = 'NDW'


def get_db():
    global _ndw_db
    if _ndw_db is None:
        _ndw_db = MetaDb(echo=False)
    return _ndw_db


def get_qm() -> QueryManager:
    return QueryManager.instance('NDW')


def create_selections():
    db = get_db()
    return {

        # DW_ORG -> DW_ORG  - 机构 
        "DW_ORG": Selection(
            db=db,
            title='机构',
            tables=['dw_org'],
            columns=[
                'ORG_NO',  # 机构号
                'ORG_NAME',  # 机构名称
                'ORG_SHORT',  # 机构简称
                'ORG_PATH',  # 机构路径
                'ORG_SCHEMA',  # 体系性质
                'UNIT_CODE',  # 单元代码
                'TREE_CODE',  # 层级码
                'IS_LEAF',  # 叶节点
                'LEVEL',  # 层级
                'UNIT_RANK',  # 规格
                'UNIT_ROLE',  # 经营定位
                'OPEN_YM',  # 设立年月
                'CLOSE_YM',  # 失效年月
                'DISABLE_FLAG',  # 失效标记
                'DIM_1',  # 统计1
                'DIM_2',  # 统计2
                'DIM_3',  # 统计3
            ]
        ),
        # DW_PDT -> DW_PDT  - 产品名录 
        "DW_PDT": Selection(
            db=db,
            title='产品名录',
            tables=['dw_pdt'],
            columns=[
                'PDT_CODE',  # 产品代号
                'PDT_NAME',  # 产品名称
                'PDT_TITLE',  # 开票品名
            ]
        ),
        # DW_PAY -> DW_PAY  - 支出 
        "DW_PAY": Selection(
            db=db,
            title='支出',
            tables=['dw_pay'],
            columns=[
                'ORG_NO',  # 机构号
                'PAY_CODE',  # 费用代码
                'YR',  # 年度
                'Y_TOTAL',  # 全年数值
                'M01',  # 1月数值
                'M02',  # 2月数值
                'M03',  # 3月数值
                'M04',  # 4月数值
                'M05',  # 5月数值
                'M06',  # 6月数值
                'M07',  # 7月数值
                'M08',  # 8月数值
                'M09',  # 9月数值
                'M10',  # 10月数值
                'M11',  # 11月数值
                'M12',  # 12月数值
            ]
        ),
        # DW_ORG_PSN -> DW_ORG_PSN  - 人员配属 
        "DW_ORG_PSN": Selection(
            db=db,
            title='人员配属',
            tables=['dw_org_psn'],
            columns=[
                'ORG_NO',  # 机构号
                'YR',  # 年度
                'MTH',  # 月份
                'EMPL_NO',  # 工号
                'DEPT_NO',  # 部门号
                'MGR_DUTY',  # 管理责任
                'JOB_PART',  # 兼岗属性
                'RES_RATIO',  # 资源配比
            ]
        ),
        # DW_DICT -> DW_DICT  - 枚举值 
        "DW_DICT": Selection(
            db=db,
            title='枚举值',
            tables=['dw_dict'],
            columns=[
                'ENUM_KEY',  # 枚举KEY
                'CODE',  # 码值
                'NAME',  # 名称
                'DISP_TEXT',  # 显示文字
                'ICON',  # 图标信息
                'DISP_ORDER',  # 显示次序
                'REMARK',  # 备注
                'DISABLE_FLAG',  # 失效标记
            ]
        ),
        # DW_PSN -> DW_PSN  - 员工 
        "DW_PSN": Selection(
            db=db,
            title='员工',
            tables=['dw_psn'],
            columns=[
                'PSN_RID',  # 员工记录ID
                'YR',  # 年度
                'MTH',  # 月份
                'EMPL_NO',  # 工号
                'EMPL_NAME',  # 姓名
                'GENDER',  # 性别
                'MOBILE_PHONE',  # 手机号码
                'EMAIL',  # 邮箱
                'ETHNIC',  # 民族
                'BIRTH_DATE',  # 出生日期
                'ORIGIN_CITY',  # 籍贯
                'ID_NUM',  # 身份证号
                'LST_EDU',  # 最后学历
                'FST_EDU',  # 第一学历
                'MAJOR',  # 专业
                'GRD_SCHOOL',  # 毕业学校
                'GRD_DATE',  # 毕业时间
                'RELATION',  # 雇佣关系
                'TRIAL',  # 试用标记
                'JOB',  # 职位
                'JOB_TITLE',  # 职位头衔
                'JOB_TYPE',  # 职种
                'JOB_LEVEL',  # 职级
                'EMPL_STATE',  # 在职状态
                'ENTRY_DATE',  # 入职日期
                'TRANS_DATE',  # 转正日期
                'QUIT_DATE',  # 离职日期
                'ONSITE',  # 用工场所
                'WORK_CITY',  # 工作地点
                'LIVING_ADDR',  # 居住地址
                'FAMILY_ADDR',  # 家庭地址
                'PAYROLL_ORG',  # 发薪单位
            ]
        ),
        # DW_SCT_PDT -> DW_SCT_PDT  - 销售明细 
        "DW_SCT_PDT": Selection(
            db=db,
            title='销售明细',
            tables=['dw_sct_pdt'],
            columns=[
                'PDT_CODE',  # 产品代号
                'SCT_ID',  # 销售合同ID
                'AMOUNT',  # 金额
            ]
        ),
        # DW_DIM_CODE -> DW_DIM_CODE  - 统计码 
        "DW_DIM_CODE": Selection(
            db=db,
            title='统计码',
            tables=['dw_dim_code'],
            columns=[
                'DM_KEY',  # 统计维度号
                'DIM_CODE',  # 统计码
                'DIM_NAME',  # 统计码名称
                'DISP_TEXT',  # 显示文字
                'DISP_ORDER',  # 显示次序
                'REMARK',  # 备注
                'CAS_CODE',  # 级联码值
                'DISABLE_FLAG',  # 失效标记
            ]
        ),
        # DW_AR -> DW_AR  - 应收款 
        "DW_AR": Selection(
            db=db,
            title='应收款',
            tables=['dw_ar'],
            columns=[
                'AR_ID',  # 应收ID
                'AR_NUM',  # 应收标识号
                'SCT_ID',  # 销售合同ID
                'CUS_NO',  # 客户号
                'BIZ_TYPE',  # 经营类别
                'AMT_TYPE',  # 款项性质
                'AMOUNT',  # 应收金额
                'REC_TERM',  # 履约条件
                'AMT_PCT',  # 占比
                'INC_FLAG',  # 回款标记
                'AR_STATE',  # 催收状态
                'AR_DATE',  # 应回款日期
                'PLN_DATE',  # 计划回款日期
                'ACT_DATE',  # 实际回款日期
                'ACT_AMOUNT',  # 已回款金额
                'PLN_DESC',  # 计划说明
                'DEMAND_DESC',  # 催收说明
                'DIM_1',  # 统计1
                'DIM_2',  # 统计2
                'DIM_3',  # 统计3
            ]
        ),
        # DW_CUS -> DW_CUS  - 客户 
        "DW_CUS": Selection(
            db=db,
            title='客户',
            tables=['dw_cus'],
            columns=[
                'CUS_NO',  # 客户号
                'CUS_NAME',  # 客户名称
                'CUS_SHORT',  # 客户简称
                'CUS_TYPE',  # 客户分类
                'ENT_NAT',  # 企业性质
                'CUS_RANK',  # 服务等级
                'RSP_ORG',  # 负责单位
                'RSP_DEPT',  # 负责部门
                'CUS_MGR',  # 客户经理
                'REGION',  # 所属区域
                'SALES_CHNL',  # 渠道
                'ADDRESS',  # 联系地址
                'CUS_CONTACT',  # 客户联系人
                'OPEN_YEAR',  # 拓展年度
                'BIZ_STATE',  # 商务状态
                'DIM_1',  # 统计1
                'DIM_2',  # 统计2
                'DIM_3',  # 统计3
                'REMARK',  # 备注
            ]
        ),
        # DW_BOP -> DW_BOP  - 商机 
        "DW_BOP": Selection(
            db=db,
            title='商机',
            tables=['dw_bop'],
            columns=[
                'BOP_ID',  # 商机ID
                'ORG_NO',  # 机构号
                'BOP_NAME',  # 商机名称
                'BOP_SHORT',  # 商机简称
                'BOP_DESC',  # 商机描述
                'LOB',  # 业务线
                'SALES_CHNL',  # 渠道
                'SALES',  # 销售人员
                'SALE_SPTS',  # 支持人员
                'CUS_NAME',  # 客户名称
                'CUS_NO',  # 客户号
                'REG_DATE',  # 登记日期
                'EST_SDATE',  # 预计启动时间
                'EXP_TIME',  # 期望上线时间
                'BOP_STATE',  # 商机状态
                'PROGRESS',  # 所处阶段
                'PRGS_DESC',  # 进展描述
                'EST_AMOUNT',  # 预估金额
                'FINAL_CUS',  # 最终用户名称
                'SOURCE',  # 商机来源
                'BUY_PTN',  # 选型方式
                'PEERS',  # 主要竞争对手
                'CMP_DESC',  # 竞争情况概述
                'NXT_ACTION',  # 行动计划
                'EST_OPPO',  # 估计签单机率
                'EST_DESC',  # 评估依据
                'DEC_MAKER',  # 决策人及其职位
                'RL_PSNS',  # 干系人
                'FIN_CONTRACTOR',  # 最终签约厂商
                'REMARK',  # 备注
                'DIM_1',  # 统计1
                'DIM_2',  # 统计2
                'DIM_3',  # 统计3
                'HBY_ITEM_ID',  # 伙伴云数据ID
            ]
        ),
        # DW_DIM -> DW_DIM  - 统计维度 
        "DW_DIM": Selection(
            db=db,
            title='统计维度',
            tables=['dw_dim'],
            columns=[
                'DM_KEY',  # 统计维度号
                'DIMENSION',  # 统计维度名称
                'CASCADE_KEY',  # 级联维度
                'REMARK',  # 备注
            ]
        ),
        # DW_DEPT -> DW_DEPT  - 部门 
        "DW_DEPT": Selection(
            db=db,
            title='部门',
            tables=['dw_dept'],
            columns=[
                'ORG_NO',  # 机构号
                'DEPT_NO',  # 部门号
                'DEPT_NAME',  # 部门名称
                'DEPT_SHORT',  # 部门简称
                'TREE_CODE',  # 层级码
                'IS_LEAF',  # 叶节点
                'DISABLE_FLAG',  # 失效标记
            ]
        ),
        # DW_INC_PLN -> DW_INC_PLN  - 收款计划 
        "DW_INC_PLN": Selection(
            db=db,
            title='收款计划',
            tables=['dw_inc_pln'],
            columns=[
                'IPLN_ID',  # 收款计划ID
                'ORG_NO',  # 机构号
                'YR',  # 年度
                'MTH',  # 月份
                'CUS_NO',  # 客户号
                'AR_TITLE',  # 款项内容
                'AMOUNT',  # 计划收款金额
                'SCT_NO',  # 销售合同号
                'AR_NUM',  # 应收标识号
                'DEPT_NO',  # 负责部门
                'OWNER',  # 负责人
                'EST_OPPO',  # 估计回款机率
                'EST_DESC',  # 估计说明
                'INC_FLAG',  # 回收结果
                'INC_AMOUNT',  # 实际收款金额
                'ACT_DESC',  # 进展说明
                'DIM_1',  # 统计1
                'DIM_2',  # 统计2
                'DIM_3',  # 统计3
            ]
        ),
        # DW_IX_BOOK -> DW_IX_BOOK  - 业绩台账 
        "DW_IX_BOOK": Selection(
            db=db,
            title='业绩台账',
            tables=['dw_ix_book'],
            columns=[
                'PM_ID',  # 台账记录ID
                'ORG_NO',  # 机构号
                'YR',  # 年度
                'IX_CODE',  # 指标代码
                'P_DATE',  # 日期
                'DIR',  # 方向
                'AMOUNT',  # 数额
                'ITEM_DESC',  # 摘要
                'REF_NO',  # 业务参考号
                'DIM_1',  # 统计1
                'DIM_2',  # 统计2
                'DIM_3',  # 统计3
                'DESC',  # 备注
            ]
        ),
        # DW_RESP -> DW_RESP  - 资源投入计划 
        "DW_RESP": Selection(
            db=db,
            title='资源投入计划',
            tables=['dw_resp'],
            columns=[
                'YR',  # 年度
                'ORG_NO',  # 机构号
                'RP_CODE',  # 资源投向编码
                'PWDAY',  # 计划资源投放人日
            ]
        ),
        # DW_RP_CODE -> DW_RP_CODE  - 资源投向 
        "DW_RP_CODE": Selection(
            db=db,
            title='资源投向',
            tables=['dw_rp_code'],
            columns=[
                'RP_CODE',  # 资源投向编码
                'RP_NAME',  # 资源投向名称
                'RP_DESC',  # 资源投向描述
                'LOB',  # 业务线
                'TREE_CODE',  # 层级码
                'DIM_1',  # 统计1
                'DIM_2',  # 统计2
                'DIM_3',  # 统计3
            ]
        ),
        # DW_WHR -> DW_WHR  - 工时 
        "DW_WHR": Selection(
            db=db,
            title='工时',
            tables=['dw_whr'],
            columns=[
                'YR',  # 年度
                'MTH',  # 月份
                'EMPL_NO',  # 工号
                'RP_CODE',  # 资源投向编码
                'WHR_DCL',  # 报工数
                'WHR_APR',  # 核准数
                'PSN_RID',  # 员工记录ID
            ]
        ),
        # DW_BFS -> DW_BFS  - 配套采购 
        "DW_BFS": Selection(
            db=db,
            title='配套采购',
            tables=['dw_bfs'],
            columns=[
                'SCT_ID',  # 销售合同ID
                'PCT_ID',  # 采购合同ID
                'SALE_AMOUNT',  # 售出金额
                'INC_FLAG',  # 回款标记
                'REMARK',  # 备注
            ]
        ),
        # DW_AP -> DW_AP  - 应付款 
        "DW_AP": Selection(
            db=db,
            title='应付款',
            tables=['dw_ap'],
            columns=[
                'AP_ID',  # 应付记录ID
                'PCT_ID',  # 采购合同ID
                'SPLR_ID',  # 供应商ID
                'AP_DATE',  # 应付日期
                'AP_AMOUNT',  # 应付金额
                'INV_NUM',  # 发票号
                'INV_STATE',  # 发票状态
                'INV_REMARK',  # 发票备注
                'CLEAR_FLAG',  # 结算标记
                'PY_ID',  # 付款记录ID
            ]
        ),
        # DW_IX_PV -> DW_IX_PV  - 指标计划值 
        "DW_IX_PV": Selection(
            db=db,
            title='指标计划值',
            tables=['dw_ix_pv'],
            columns=[
                'IX_CODE',  # 指标代码
                'ORG_NO',  # 机构号
                'YR',  # 年度
                'Y_TOTAL',  # 全年数值
                'M01',  # 1月数值
                'M02',  # 2月数值
                'M03',  # 3月数值
                'M04',  # 4月数值
                'M05',  # 5月数值
                'M06',  # 6月数值
                'M07',  # 7月数值
                'M08',  # 8月数值
                'M09',  # 9月数值
                'M10',  # 10月数值
                'M11',  # 11月数值
                'M12',  # 12月数值
                'Q1',  # 1季度数值
                'Q2',  # 2季度数值
                'Q3',  # 3季度数值
                'Q4',  # 4季度数值
                'H1',  # 上半年数
                'H2',  # 下半年数
                'MY01',  # 1月累计
                'MY02',  # 2月累计
                'MY03',  # 3月累计
                'MY04',  # 4月累计
                'MY05',  # 5月累计
                'MY06',  # 6月累计
                'MY07',  # 7月累计
                'MY08',  # 8月累计
                'MY09',  # 9月累计
                'MY10',  # 10月累计
                'MY11',  # 11月累计
                'MY12',  # 12月累计
            ]
        ),
        # DW_AR_MATCH -> DW_AR_MATCH  - 应收对账 
        "DW_AR_MATCH": Selection(
            db=db,
            title='应收对账',
            tables=['dw_ar_match'],
            columns=[
                'INC_ID',  # 收款记录ID
                'AR_ID',  # 应收ID
                'INC_AMOUNT',  # 回款金额
            ]
        ),
        # DW_BPAY -> DW_BPAY  - 采购付款记录 
        "DW_BPAY": Selection(
            db=db,
            title='采购付款记录',
            tables=['dw_bpay'],
            columns=[
                'PY_ID',  # 付款记录ID
                'PCT_ID',  # 采购合同ID
                'PDATE',  # 付款日期
                'AMOUNT',  # 付款金额
                'ITEM_DESC',  # 摘要
                'MATCH_FLAG',  # 银行对账标识
                'VCH_NUM',  # 记账凭证号
                'BOOK_AMOUNT',  # 记账金额
                'REMARK',  # 备注
                'DIM_1',  # 统计1
                'DIM_2',  # 统计2
                'DIM_3',  # 统计3
            ]
        ),
        # DW_IX -> DW_IX  - 指标项 
        "DW_IX": Selection(
            db=db,
            title='指标项',
            tables=['dw_ix'],
            columns=[
                'IX_CODE',  # 指标代码
                'IX_NAME',  # 指标名称
                'DISP_TEXT',  # 显示文字
                'IX_DESC',  # 指标描述
                'SUM_DIMS',  # 汇总维度
                'IX_EXPR',  # 计算表达式
                'DISP_ORDER',  # 显示次序
                'CALC_BATCH',  # 计算批次
                'DISABLE_FLAG',  # 失效标记
            ]
        ),
        # DW_PSN_MTH -> DW_PSN_MTH  - 人月数 
        "DW_PSN_MTH": Selection(
            db=db,
            title='人月数',
            tables=['dw_psn_mth'],
            columns=[
                'ORG_NO',  # 机构号
                'YR',  # 年度
                'MTH',  # 月份
                'PM_TOTAL',  # 全口径人月
                'PM_STD',  # 可比口径人月
                'PM_REGULAR',  # 常规人月
                'PM_INTERN',  # 实习生人月
                'PM_OUTSRC',  # 外包人月
                'PM_ONSITE',  # 驻场人月
            ]
        ),
        # DW_IX_FV -> DW_IX_FV  - 指标执行值 
        "DW_IX_FV": Selection(
            db=db,
            title='指标执行值',
            tables=['dw_ix_fv'],
            columns=[
                'IX_CODE',  # 指标代码
                'ORG_NO',  # 机构号
                'YR',  # 年度
                'Y_TOTAL',  # 全年数值
                'M01',  # 1月数值
                'M02',  # 2月数值
                'M03',  # 3月数值
                'M04',  # 4月数值
                'M05',  # 5月数值
                'M06',  # 6月数值
                'M07',  # 7月数值
                'M08',  # 8月数值
                'M09',  # 9月数值
                'M10',  # 10月数值
                'M11',  # 11月数值
                'M12',  # 12月数值
                'Q1',  # 1季度数值
                'Q2',  # 2季度数值
                'Q3',  # 3季度数值
                'Q4',  # 4季度数值
                'H1',  # 上半年数
                'H2',  # 下半年数
                'MY01',  # 1月累计
                'MY02',  # 2月累计
                'MY03',  # 3月累计
                'MY04',  # 4月累计
                'MY05',  # 5月累计
                'MY06',  # 6月累计
                'MY07',  # 7月累计
                'MY08',  # 8月累计
                'MY09',  # 9月累计
                'MY10',  # 10月累计
                'MY11',  # 11月累计
                'MY12',  # 12月累计
            ]
        ),
        # DW_SCT -> DW_SCT  - 销售合同 
        "DW_SCT": Selection(
            db=db,
            title='销售合同',
            tables=['dw_sct'],
            columns=[
                'SCT_ID',  # 销售合同ID
                'CUS_NO',  # 客户号
                'SCT_NO',  # 销售合同号
                'SCT_TITLE',  # 合同名称
                'AMOUNT',  # 合同金额
                'BUY_AMOUNT',  # 代购金额
                'LOB',  # 业务线
                'BIZ_TYPE',  # 经营类别
                'ISSUE_YEAR',  # 签约年份
                'ISSUE_DATE',  # 签约日期
                'SDATE',  # 开始日期
                'EDATE',  # 到期日期
                'OP_SCT_NO',  # 对方合同号
                'SCT_STATE',  # 合同状态
                'TERMS',  # 主要条款
                'ORG_NOM',  # 署名单位
                'ORG_SALE',  # 签约单位
                'DEPT_SALE',  # 签约部门
                'ORG_EXEC',  # 履约单位
                'DEPT_EXEC',  # 履约部门
                'ORG_BOOK',  # 核算单位
                'BIZ_OWNER',  # 商务经办人
                'PFM_OWNER',  # 业绩归属人
                'DOC_OWNER',  # 档案负责人
                'DOC_STATE',  # 归档状态
                'REMARK',  # 备注
                'DIM_1',  # 统计1
                'DIM_2',  # 统计2
                'DIM_3',  # 统计3
                'PA_NAME',  # 甲方署名单位
                'ACT_USER',  # 实际用户
            ]
        ),
        # DW_SPLR -> DW_SPLR  - 供应商 
        "DW_SPLR": Selection(
            db=db,
            title='供应商',
            tables=['dw_splr'],
            columns=[
                'SPLR_ID',  # 供应商ID
                'SPLR_NO',  # 供应商编号
                'SPLR_NAME',  # 供应商名称
                'SPLR_SHORT',  # 供应商简称
                'SPLR_TYPE',  # 供应商分类
                'SPLR_STATE',  # 合作状态
                'FST_YEAR',  # 开始合作年份
                'PUR_TYPE',  # 采购分类
                'PUR_PDT',  # 主要采购产品
                'STD_PRICE',  # 基准价格
                'PUR_TERMS',  # 商务条件
                'SPLR_RANK',  # 供应商评级
                'BANK',  # 开户行
                'ACCOUNT_NO',  # 账号
                'ADDR',  # 地址
                'TEL',  # 联系电话
                'SPLR_CONTACT',  # 联系人
                'REMARK',  # 备注
                'DIM_1',  # 统计1
                'DIM_2',  # 统计2
                'DIM_3',  # 统计3
            ]
        ),
        # DW_ENUM -> DW_ENUM  - 枚举类型 
        "DW_ENUM": Selection(
            db=db,
            title='枚举类型',
            tables=['dw_enum'],
            columns=[
                'ENUM_KEY',  # 枚举KEY
                'ENUM_NAME',  # 枚举名称
                'V_TYPE',  # 值类型
                'REMARK',  # 备注
            ]
        ),
        # DW_PCT -> DW_PCT  - 采购合同 
        "DW_PCT": Selection(
            db=db,
            title='采购合同',
            tables=['dw_pct'],
            columns=[
                'PCT_ID',  # 采购合同ID
                'SPLR_ID',  # 供应商ID
                'PCT_NO',  # 采购合同号
                'PCT_TITLE',  # 采购合同名称
                'PUR_TYPE',  # 采购分类
                'PUR_PDT',  # 主要采购产品
                'AMOUNT',  # 合同金额
                'ISSUE_YEAR',  # 签约年份
                'ISSUE_DATE',  # 签约日期
                'SDATE',  # 开始日期
                'EDATE',  # 到期日期
                'OP_SCT_NO',  # 对方合同号
                'PCT_STATE',  # 采购合同状态
                'TERMS',  # 主要条款
                'ORG_NOM',  # 署名单位
                'ORG_SALE',  # 签约单位
                'DEPT_SALE',  # 签约部门
                'ORG_EXEC',  # 履约单位
                'DEPT_EXEC',  # 履约部门
                'ORG_BOOK',  # 核算单位
                'BIZ_OWNER',  # 商务经办人
                'DOC_OWNER',  # 档案负责人
                'DOC_STATE',  # 归档状态
                'REMARK',  # 备注
                'DIM_1',  # 统计1
                'DIM_2',  # 统计2
                'DIM_3',  # 统计3
            ]
        ),
        # DW_BOP_TRACE -> DW_BOP_TRACE  - 商机跟进 
        "DW_BOP_TRACE": Selection(
            db=db,
            title='商机跟进',
            tables=['dw_bop_trace'],
            columns=[
                'TR_ID',  # 商机跟进ID
                'BOP_ID',  # 商机ID
                'TRC_DATE',  # 跟进日期
                'TRC_MAN',  # 跟进人员
                'COMM_TYPE',  # 客户沟通方式
                'TRC_DESC',  # 情况概述
                'BOP_DESC',  # 详情
                'HBY_ITEM_ID',  # 伙伴云数据ID
                'HBY_BOP_ITEM_ID',  # 伙伴云商机数据ID
            ]
        ),
        # DW_INC -> DW_INC  - 收款记录 
        "DW_INC": Selection(
            db=db,
            title='收款记录',
            tables=['dw_inc'],
            columns=[
                'INC_ID',  # 收款记录ID
                'CUS_NO',  # 客户号
                'SCT_ID',  # 销售合同ID
                'INC_DATE',  # 收款日期
                'AMOUNT',  # 收款金额
                'ITEM_DESC',  # 摘要
                'INV_STATE',  # 发票状态
                'INV_NUM',  # 发票号
                'MATCH_FLAG',  # 银行对账标识
                'VCH_NUM',  # 记账凭证号
                'BOOK_AMOUNT',  # 记账金额
                'REMARK',  # 收款备注
                'DIM_1',  # 统计1
                'DIM_2',  # 统计2
                'DIM_3',  # 统计3
            ]
        ),
        # DW_DOC -> DW_DOC  - 合同档案 
        "DW_DOC": Selection(
            db=db,
            title='合同档案',
            tables=['dw_doc'],
            columns=[
                'DR_ID',  # 档案记录ID
                'SCT_ID',  # 销售合同ID
                'PCT_ID',  # 采购合同ID
                'DOC_FLAG',  # 分类标记
                'DOC_LABEL',  # 档案标签
                'DOC_CONTENT',  # 档案内容
            ]
        ),
    }



# -*- logics.py: python ; coding: utf-8 -*-
import os
from decimal import Decimal

import pandas as pd
from pandas import DataFrame
from sqlalchemy import literal
from sqlalchemy.sql.functions import func

from yus import UnitType, ChooseMode, DisplayAttitude
from yui.ydict import DictLoader, DictFetcher
from yus import Op, QueryR, ColumnSpec
from yus.dbm import extract_qc, Selection, MetaDb
from yus_demo_models import get_qm, create_selections, get_db, NAME_SPACE
from yut import m64_dec, safe_div, Obj, get_attr


def show_progress(text, *args):
    print('*')
    print('*' * 100)
    print(f'* {text}\t', *args)
    print('*' * 100)
    print('*')


class NDW_DictLoader(DictLoader):

    def __init__(self):
        super().__init__(NAME_SPACE)

    def load_enum_dict(self, enum_key) -> dict:
        if enum_key == '$enum':  # 返回枚举类型列表
            df = get_qm().get_selection('DW_ENUM').query().dataframe()
            return {r.ENUM_KEY: r.ENUM_NAME for r in df.itertuples()}
        else:
            df = get_qm().get_selection('DW_DICT').query(ENUM_KEY=enum_key, DISABLE_FLAG=0).dataframe()
            return {r.CODE: r.DISP_TEXT for r in df.itertuples()}

    def load_dim_dict(self, dim_key) -> dict:
        if dim_key == '$dim':  # 返回统计类型列表
            df = get_qm().get_selection('DW_DIM').query().dataframe()
            return {r.DM_KEY: r.DIMENSION for r in df.itertuples()}
        else:
            df = get_qm().get_selection('DW_DIM_CODE').query(DM_KEY=dim_key, DISABLE_FLAG=0).dataframe()
            return {r.DIM_CODE: r.DISP_TEXT for r in df.itertuples()}


_default_dict_loader = NDW_DictLoader()


def init_env():
    show_progress('[yus_demo.init_env]', '初始化运行环境')
    env_key = 'YUS_DATABASE_URL'
    MetaDb.DEFAULT_DATABASE_URI = m64_dec(os.environ.get(env_key))
    print(f' -- 从系统变量 {env_key} 获取数据库连接URI:', MetaDb.DEFAULT_DATABASE_URI)
    for k, v in create_selections().items():
        get_qm().register_selection(k, v)
    print(f'注册默认的Selection,共 {get_qm().selection_count()} 个.')
    DictFetcher.instance().set_loader(_default_dict_loader)


def load_orgs(schema) -> list[Obj]:
    df_org = get_qm().get_selection('DW_ORG').query(ORG_SCHEMA=schema).dataframe()
    return [Obj(ORG_NO=r.ORG_NO,
                ORG_NAME=r.ORG_NAME,
                ORG_SHORT=r.ORG_SHORT,
                ORG_PATH=r.ORG_PATH,
                ORG_SCHEMA=r.ORG_SCHEMA,
                UNIT_CODE=r.UNIT_CODE,
                LEVEL=r.LEVEL,
                TREE_CODE=r.TREE_CODE,
                IS_LEAF=r.IS_LEAF,
                UNIT_RANK=r.UNIT_RANK,
                UNIT_ROLE=r.UNIT_ROLE,
                ) for r in df_org.itertuples()]


def simple_query():
    show_progress('[yus_demo.simple_query]', '单表查询，简单条件')
    yr, mth, schema = 2025, 5, 1
    print(f'{yr}年{mth}月组织机构人数统计(按业务体系):')
    df_org = get_qm().get_selection('DW_ORG').query(ORG_SCHEMA=schema).dataframe()
    org_psn = get_qm().get_selection('DW_ORG_PSN')
    for r in df_org.itertuples():
        df_psn = org_psn.query(ORG_NO=r.ORG_NO, YR=2025, MTH=5)
        print(f'{r.ORG_NO}-{r.ORG_SHORT:30}{r.ORG_PATH:<20}:\t {df_psn.dataframe().shape[0]}')


def group_month(df, col_date, col_amount, col_groups: list, fn_mth_col=lambda m: f"M{m:02d}"):
    """
    将按照日期纵向排列的数据表转为按照月份横向排列的数据表.
    :param df: 源表
    :param col_date: 日期列名称
    :param col_amount: 金额/数值列名称
    :param col_groups: 分组列集合（日期之外的其它分组列名）
    :param fn_mth_col: 月份列名转换函数
    :return: 包含col_groups以及1-12月列名的横表
    """
    temp_col_name = '_MTH_'
    # 确保日期列是datetime类型
    df[col_date] = pd.to_datetime(df[col_date])
    # 增加月份临时列
    df[temp_col_name] = df[col_date].dt.month

    # 按col_groups和_MTH_分组，计算每月金额
    col_groups.append(temp_col_name)
    df_m = df.groupby(col_groups)[col_amount].sum().unstack()
    df_m = df_m.fillna(0)

    # 重命名月份列
    df_m.columns = [fn_mth_col(col) for col in df_m.columns]

    # 确保所有月份都存在（1-12月），如果某月没有数据则填充为0
    for month in range(1, 13):
        col_name = fn_mth_col(month)
        if col_name not in df_m.columns:
            df_m[col_name] = 0

    # 重新排序列，确保顺序正确（1月到12月）
    df_m = df_m[[fn_mth_col(m) for m in range(1, 13)]]

    # 重置索引，使col_groups成为列
    df_m = df_m.reset_index()

    return df_m


def income_report(params) -> QueryR:
    def m_col(m):
        return f'M{m:02d}'

    def new_row(df_source, disp_ord, tag) -> DataFrame:
        df_target = df_source.copy()
        df_target['ORD'] = disp_ord
        df_target['TAG'] = tag
        return df_target

    def percent(va, vb):
        ret = safe_div(va, vb, none_as=0)
        return ret

    year = get_attr(extract_qc(params, 'YEAR'), 'value')
    orgs = get_attr(extract_qc(params, 'ORG_EXEC'), 'value')
    biz_types = get_attr(extract_qc(params, 'BIZ_TYPE'), 'value')
    units = get_attr(extract_qc(params, 'UNITS'), 'value', 10000)

    show_progress('[yus_demo.income_report]', f'年度回款统计: year={year};orgs={orgs};biz_types={biz_types}')
    ss_inc = Selection(db=get_db(),
                       title='年度回款统计',
                       tables=['dw_inc', 'dw_sct'],
                       columns=['a.INC_DATE',
                                'b.ORG_EXEC' if orgs else literal("10000").label('ORG_EXEC'),
                                lambda ss: func.sum(ss.a.AMOUNT / units).label('AMOUNT'),
                                ],
                       joins=[('a.SCT_ID', 'b.SCT_ID')],
                       unit_spec={
                           'BIZ_TYPE': {'utype': UnitType.CHOOSE, 'mode': ChooseMode.LIST,
                                        'dict_value': "{0:'自营',1:'代购',2:'驻场'}"},
                           'AMOUNT': {'utype': UnitType.CURRENCY},
                       },
                       group_bys=['a.INC_DATE', 'b.ORG_EXEC'],
                       order_bys=['b.ORG_EXEC'],
                       )

    qm = get_qm()
    qc = {'INC_DATE': [Op.BETWEEN, (f'{year}-01-01', f'{year}-12-31')]}
    # 查询条件
    if orgs:
        qc.update({'b.ORG_EXEC': [Op.IN, orgs]})
    if biz_types:
        qc.update({'b.BIZ_TYPE': [Op.IN, biz_types]})

    df = qm.query_selection(selection=ss_inc, condition=qc).dataframe()
    # print(df.to_string())

    df = group_month(df, col_date='INC_DATE', col_amount='AMOUNT', col_groups=['ORG_EXEC'], fn_mth_col=m_col)

    # 4行：当月、当月累计、当月占全年、月累计占全年
    df_m = new_row(df, 0, 'INC')
    df_y = new_row(df_m, 10, 'INC.Y')
    df_p = new_row(df_m, 20, 'INC.P')
    df_yp = new_row(df_m, 30, 'INC.YP')

    # 逐月计算累计值
    for month in range(1, 13):
        cm, pm = m_col(month), m_col(month - 1)
        if month == 1:
            df_y[cm] = df_y[cm]  # M01累计=M01
        else:
            df_y[cm] = df_y[pm] + df_m[cm]
    # 计算占比
    for month in range(1, 13):
        cm, ym = m_col(month), m_col(12)
        df_p[cm] = percent(df_m[cm], df_y[ym]) * 100
        df_yp[cm] = percent(df_y[cm], df_y[ym]) * 100

    # 合并各行
    result = pd.concat([df_m, df_y, df_p, df_yp], ignore_index=True)
    # 排序,年累计行会排在后面
    result = result.sort_values(['ORG_EXEC', 'ORD'])
    column_specs = [
        ColumnSpec(name='ORG_EXEC', comment='业务单元', utype=UnitType.LOOKUP, lookup_key='yut_demo.OrgLookup',
                   format='{name}'),

    ]
    column_specs.extend(
        [ColumnSpec(name=m_col(m), comment=f'{m}月', utype=UnitType.CURRENCY) for m in range(1, 13)])
    column_specs.extend([ColumnSpec(name='ORD', comment='显示次序', utype=UnitType.INT, ),
                         ColumnSpec(name='TAG', comment='项目', utype=UnitType.CHOOSE,
                                    dict_value="{'INC':'月回款','INC.Y':'累计回款','INC.P':'月占比%','INC.YP':'累计占比%'}",
                                    format='{name}'), ])
    return QueryR(result, column_specs)


def selection_sct_list() -> Selection:
    show_progress('[yus_demo.selection_sct_list]',
                  '复杂查询：合同列表，包含已回款金额、余额的计算以及合同与客户的引用关系')
    # 复杂查询：合同列表，包含已回款金额、余额的计算以及合同与客户的引用关系
    sct_tables = ['dw_sct', 'dw_cus', 'dw_inc']
    sct_columns = [
        'SCT_ID',  # 销售合同ID
        'CUS_NO',  # 客户号
        'b.CUS_SHORT',  # 客户简称
        'b.CUS_TYPE',  # 客户类型
        'SCT_NO',  # 销售合同号
        'SCT_TITLE',  # 合同名称
        'AMOUNT',  # 合同金额
        'BUY_AMOUNT',  # 代购金额
        'LOB',  # 业务线
        'BIZ_TYPE',  # 经营类别
        'ISSUE_YEAR',  # 签约年份
        'ISSUE_DATE',  # 签约日期
        'SDATE',  # 开始日期
        'EDATE',  # 到期日期
        'OP_SCT_NO',  # 对方合同号
        'SCT_STATE',  # 合同状态
        # 'TERMS',  # 主要条款
        'ORG_NOM',  # 署名单位
        'ORG_SALE',  # 签约单位
        # 'DEPT_SALE',  # 签约部门
        'ORG_EXEC',  # 履约单位
        # 'DEPT_EXEC',  # 履约部门
        'ORG_BOOK',  # 核算单位
        # 'BIZ_OWNER',  # 商务经办人
        # 'PFM_OWNER',  # 业绩归属人
        # 'DOC_OWNER',  # 档案负责人
        # 'DOC_STATE',  # 归档状态
        # 'REMARK',  # 备注
        # 'DIM_1',  # 统计1
        # 'DIM_2',  # 统计2
        # 'DIM_3',  # 统计3
        # 'PA_NAME',  # 甲方署名单位
        # 'ACT_USER',  # 实际用户
        lambda qs: func.sum(qs.f('c.AMOUNT')).label('INC_AMOUNT'),  # 已回款金额
    ]
    sct_joins = [
        ('a.CUS_NO', 'b.CUS_NO'),
        ('a.SCT_ID', 'c.SCT_ID', True),  # 与回款表必须使用外连接
    ]
    sct_ext_comments = {
        'CUS_NO': '客户号',
        'CUS_SHORT': '客户简称',
        'CUS_TYPE': '客户类型',
        'INC_AMOUNT': '已回款金额',
        'BALANCE': '合同余额',
    }
    sct_unit_spec = {
        'BIZ_TYPE': {'utype': UnitType.CHOOSE, 'mode': ChooseMode.LIST,
                     'dict_value': "{0:'自营',1:'代购',2:'驻场'}",
                     'attitude': {DisplayAttitude.WEAK_NEGATIVE: ['1'],
                                  DisplayAttitude.WEAK_POSITIVE: ['2'],
                                  # AttitudeToken.POSITIVE: ['0'],
                                  }
                     },
        'SCT_ID': {'utype': UnitType.INT, 'leading_zero': 6,
                   'link_to': 'yut_demo.show_contract({_value}, False)'},
        'CUS_ID': {'utype': UnitType.INT, 'format': '%08d',
                   'link_to': 'nmis.app.show_custom({value},{CUST_NO},{CUST_NAME})'},
        'CUS_TYPE': {'utype': UnitType.CHOOSE, 'mode': ChooseMode.LIST, 'dim_key': 'CUS.TYPE', },
        'SCT_NO': {'utype': UnitType.TEXT,
                   'link_to': 'yut_demo.show_contract({_value}, True)'},
        'INC_AMOUNT': {'utype': UnitType.CURRENCY},
        'BALANCE': {'utype': UnitType.CURRENCY, 'attitude_color': True},
        'ISSUE_YEAR': {'utype': UnitType.YEAR},
        'LOB': {'utype': UnitType.CHOOSE, 'mode': ChooseMode.LIST, 'dim_key': 'BIZ.LOB', },
        'SCT_STATE': {'utype': UnitType.CHOOSE, 'mode': ChooseMode.LIST, 'enum_key': 'SCT_STATE',
                      'attitude': {DisplayAttitude.NEGATIVE: [9],
                                   DisplayAttitude.WEAK_POSITIVE: [3],
                                   DisplayAttitude.WEAK_NEGATIVE: [2],
                                   DisplayAttitude.POSITIVE: [4],
                                   }},
        'INC_RATE': {'utype': UnitType.PERCENT},
        'ORG_NOM': {'utype': UnitType.LOOKUP, 'lookup_key': 'yut_demo.OrgLookup', 'multi_value': False},
        'ORG_SALE': {'utype': UnitType.LOOKUP, 'lookup_key': 'yut_demo.OrgLookup', 'multi_value': False},
        'ORG_EXEC': {'utype': UnitType.LOOKUP, 'lookup_key': 'yut_demo.OrgLookup', 'multi_value': False},
        'ORG_BOOK': {'utype': UnitType.LOOKUP, 'lookup_key': 'yut_demo.OrgLookup', 'multi_value': False},

    }

    def sct_post_process(sel, df):
        df['BALANCE'] = df.AMOUNT - df.INC_AMOUNT.fillna(Decimal(0.0))  # 后处理方法增加余额列
        df['INC_RATE'] = df.apply(
            lambda x: Decimal(safe_div(x.INC_AMOUNT, x.AMOUNT, zero_div=0.0, none_as=0.0)) * Decimal(100.0),
            axis=1)  # 后处理方法增加回款百分比
        sel.add_ext_comments(INC_RATE='回款比例')  # 可以在后处理中增加列附注INC_RATE，也可以在创建Selection时由ext_comments预先指定
        # 以下处理附带的查询条件
        for attr in ['BALANCE', 'INC_RATE']:
            c = sel.get_condition(attr)
            if c:
                df = c.op.query_df(df, c.attr, c.value)
        df['INC_AMOUNT'] = df['INC_AMOUNT'].fillna(Decimal(0.0))  # 将NaN视为0
        return df

    def q_maker(q, s, *args, **kwargs):
        q = s.add_filter(q, s.a.ISSUE_YEAR, kwargs['year'], Op.IN)
        q = s.add_filter(q, 'a.BIZ_TYPE', kwargs['biz_type'], Op.LIKE)
        q = s.add_filter(q, 'a.AMOUNT', kwargs['amount'], Op.BETWEEN)
        q = q.order_by(s.a.CUS_NO, s.a.SCT_NO)
        return q

    ss_sct = Selection(
        db=get_db(),
        title='销售合同',
        comment='带大字段的销售合同',
        tables=sct_tables,
        columns=sct_columns,
        joins=sct_joins,
        ext_comments=sct_ext_comments,
        unit_spec=sct_unit_spec,
        group_bys=['a.SCT_ID'],
        order_bys=['a.SCT_NO'],
        make_q=q_maker,
        post_process=sct_post_process,
    )
    return ss_sct


def init_models():
    get_qm().register_selection('sct_list', 'yus_demo.selection_sct_list()')


def query_sct_list(qc) -> QueryR:
    json = get_qm().query_selection(ds_key='sct_list', condition=qc).encode()
    return QueryR.decode(json)


def generate_model():
    show_progress('[yus_demo.generate_model]', '从数据库反射生成model文件')
    meta_db = MetaDb(echo=True)
    meta_db.generate_code(py_filename=r'yus_demo_models.py',
                          save_to_file=True,
                          name_pattern='[dD][wW]_.',
                          name_space='NDW')


if __name__ == '__main__':
    init_env()
    # generate_model()
    init_models()
    # simple_query()
    # for org in load_orgs(1):
    #     print(org)
    qr = income_report({'YEAR': (Op.EQU, 2024),
                        'ORG_EXEC': (Op.IN, ['10002', '10003', '10007']),
                        })
    print(qr.dataframe().to_string())
    qr = query_sct_list({'ISSUE_YEAR': (Op.IN, (2024, 2025)),
                         'BIZ_TYPE': (Op.LIKE, '%'),
                         'AMOUNT': (Op.GT, 0.1),
                         })
    print(qr.dataframe().to_string())



from contextlib import contextmanager
from typing import Generator

from loguru import logger
from sqlalchemy import Column, Integer, String
from sqlalchemy.orm import declarative_base, sessionmaker
from sqlalchemy.engine import Engine

import yus_demo
from ndw import get_db, get_qm


@contextmanager
def db_session(engine: Engine) -> Generator:
    """上下文管理器用于自动处理数据库会话"""
    Session = sessionmaker(bind=engine)
    session = Session()
    try:
        yield session
        session.commit()
    except Exception:
        session.rollback()
        raise
    finally:
        session.close()


def lazy_query(session, query, batch_size=100) -> Generator:
    """
    使用 SQLAlchemy 的生成器懒加载数据库结果

    :param session: SQLAlchemy 会话对象
    :param query: SQLAlchemy 查询对象 (Select 或 ORM 查询)
    :param batch_size: 每次从数据库获取的批处理大小
    """
    # 创建可滚动的服务器端游标
    result = session.execute(query.execution_options(yield_per=batch_size))

    # 逐批获取结果
    while True:
        # 获取下一批结果
        batch = result.fetchmany(batch_size)
        if not batch:
            break  # 没有更多数据时退出循环

        # 逐行生成结果
        for row in batch:
            # 如果是 ORM 查询，返回实体对象；如果是核心查询，返回行对象
            yield row[0] if isinstance(row, tuple) and hasattr(row[0], '__table__') else row


# 示例使用
if __name__ == "__main__":
    yus_demo.init_env()
    engine = get_db().engine

    # 使用Lazy Load
    with db_session(engine) as session:
        qs = get_qm().get_selection('DW_AR')
        # 创建查询
        core_query = qs._create_query()

        # 获取懒加载生成器
        core_generator = lazy_query(session, core_query, batch_size=100)

        logger.info("Processing using SQL Core expressions:")
        for idx, row in enumerate(core_generator, 1):
            logger.info(f"{idx:>6d}: [#{row.AR_ID:06d}] {row.AR_NUM:<30s} - {row.SCT_ID:<10d} {row.AMOUNT:>20,.2f}")

            # 模拟提前退出
            if idx == 500:
                print("Stopped after 150 records")
                break

    # 使用普通查询
    qs = get_qm().get_selection('DW_AR')
    # 创建查询
    q = qs._create_query()
    logger.info("使用普通查询:")
    for idx, row in enumerate(q.all()):
        logger.info(f"{idx:>6d}: [#{row.AR_ID:06d}] {row.AR_NUM:<30s} - {row.SCT_ID:<10d} {row.AMOUNT:>20,.2f}")

        # 模拟提前退出
        if idx == 500:
            print("Stopped after 150 records")
            break



# ytree.py - 通用树形显示控件
from typing import Any

from PySide6.QtCore import Qt, QSize, Signal, QModelIndex
from PySide6.QtWidgets import (
    QTreeView, QAbstractItemView
)

from .ymodel import YTreeModel


class YTreeView(QTreeView):
    """
    支持YTreeModel的树形组件
    """
    itemChanged = Signal(QModelIndex, object)

    def __init__(self, parent=None,
                 model: YTreeModel = None,
                 hidden_columns=None,
                 show_grid=False,
                 ):
        super().__init__(parent)
        self.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.setIconSize(QSize(12, 12))
        self.setUniformRowHeights(True)
        self.hidden_columns = hidden_columns if hidden_columns is not None else []
        self._model = None
        if model:
            self.setup_with_model(model)
        self.header().setDefaultAlignment(Qt.AlignmentFlag.AlignCenter)
        if show_grid:
            self.set_grid()

    def item_changed(self):
        index = self.currentIndex()
        if self._model:
            row_dat = self._model.data(index, Qt.ItemDataRole.UserRole)
            self.itemChanged.emit(index, row_dat)

    def setup_with_model(self, model: YTreeModel) -> None:
        self.setModel(model)
        col_index = 0
        for col in self._model.columns:
            if col in self.hidden_columns:
                continue
            self.setColumnWidth(col_index, 100)
            col_index = col_index + 1

    def setModel(self, model, /):
        self._model = model
        super().setModel(model)
        if model and model.rowCount() > 0:
            self.selectionModel().selectionChanged.connect(self.item_changed)

    def get_selected_data(self, role=Qt.ItemDataRole.UserRole) -> Any:
        indexes = self.selectedIndexes()
        return indexes[0].data(role) if indexes else None

    def set_columns_width(self, columns_width: list):
        if not columns_width:
            return
        d = columns_width if columns_width is dict else {i: c for i, c in enumerate(columns_width)}
        for c, w in d.items():
            c_idx = c if type(c) is int else self._model.columns.index(c)
            if 0 <= c_idx < len(self._model.columns):
                self.setColumnWidth(c_idx, w)

    def set_grid(self):
        self.setStyleSheet("""
            QTreeView {
                show-decoration-selected: 0;
                gridline-color: palette(Midlight);
            }
            QTreeView::item {
                border: 1px solid palette(Midlight);
                border-top-color: transparent;
                border-left-color: transparent;
                border-right-color: palette(Midlight);
                border-bottom-color: palette(Midlight);
            }
            QTreeView::item:selected {
                background-color: palette(Highlight);
                color: palette(HighlightedText);
            }
        """)

        """
        QHeaderView::section
        {
            background-color:  # f0f0f0;
            border: 1px solid  # c0c0c0;
        }
        """

    def expand_root(self):
        model = self.model()
        root_index = self.rootIndex()  # 获取根索引
        for i in range(model.rowCount(root_index)):
            child_index = model.index(i, 0, root_index)
            self.expand(child_index)  # 展开一级节点



# -*- wdg.py: python ; coding: utf-8 -*-
import sys
from datetime import datetime, timedelta
from decimal import Decimal
from typing import Union, Any

from PySide6 import QtWidgets
from PySide6.QtCore import Qt, QObject, Signal, QEvent, QSize, Slot, QDate, QRectF
from PySide6.QtGui import QEnterEvent, QPalette, QAction, QFont, QColor, QPainter, QPen
from PySide6.QtWidgets import QWidget, QStackedWidget, QFrame, QVBoxLayout, QToolButton, QHBoxLayout, QGridLayout, \
    QTabWidget, QSizePolicy, QMenu, QSpacerItem, QPushButton, QLabel, QSpinBox, QLayout, QButtonGroup, QProgressBar, \
    QScrollArea, QApplication, QTableWidget, QAbstractItemView, QHeaderView, QTableWidgetItem

import yui
from yui import img, create_layout, create_layout_widget, create_button, change_font, create_menu_button, \
    create_date_edit, load_icon, create_widget, create_toolbar, create_tool_button, pop_question, colored_pixmap, \
    set_label_color, get_mainwindow
from yut import to_date, Obj, copy_obj, format_date, is_collection, get_attr, set_attr, to_float


class YTitle(QWidget):
    """
    通用标题组件，包含图标和文字

    参数:
        parent: 父组件
        icon_path: 图标路径
        text: 标题文本
        icon_size: 图标大小 (默认 24x24)
        font_size: 文本字体大小 (默认14，加粗)
        spacing: 图标和文本间距 (默认10)
    """

    def __init__(self, text, parent=None,
                 icon_path=img.P_4_DOTS,
                 icon_size=QSize(24, 24),
                 on_click=None,
                 click_param=None,
                 alignment=Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter,
                 font_size=14,
                 font_bold=True,
                 spacing=4,
                 object_name=None):
        super().__init__(parent)

        self.on_click = on_click
        self.click_param = click_param
        self.setObjectName(object_name)

        # 初始化布局
        self.layout = QHBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(spacing)

        # 图标标签
        self.icon_label = QLabel(self)
        self.icon_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.set_icon(icon_path, icon_size)

        # 文本标签
        self.text_label = QLabel(self)
        self.text_label.setTextFormat(Qt.TextFormat.RichText)
        self.text_label.setTextInteractionFlags(Qt.TextInteractionFlag.TextBrowserInteraction)
        self.text_label.setOpenExternalLinks(False)  # 禁用自动打开
        self.text_label.linkActivated.connect(self.link_activated)
        self.set_text(text)

        # 设置字体
        font = QFont()
        font.setPointSize(font_size)
        font.setBold(font_bold)
        self.set_font(font)

        # 添加组件到布局
        self.layout.addWidget(self.icon_label)
        self.layout.addWidget(self.text_label, 1, alignment=alignment)
        self.layout.addStretch(0)

    def link_activated(self, link):
        if self.on_click:
            self.on_click(self, link)

    def set_icon(self, icon_path, size=None):
        if icon_path:
            pixmap = colored_pixmap(icon_path)
            if size:
                pixmap = pixmap.scaled(size, Qt.AspectRatioMode.KeepAspectRatio,
                                       Qt.TransformationMode.SmoothTransformation)
            self.icon_label.setPixmap(pixmap)
        else:
            self.icon_label.clear()

    def set_text(self, text):
        href = self.click_param if self.click_param else '#'
        text = f"<a href='{href}' style='text-decoration: none; color:#c0c0c0'>{text}</a>" if self.on_click else text
        self.text_label.setText(text)

    def set_font(self, font):
        self.text_label.setFont(font)

    def set_icon_size(self, size):
        if self.icon_label.pixmap():
            pixmap = self.icon_label.pixmap()
            self.icon_label.setPixmap(pixmap.scaled(size, Qt.AspectRatioMode.KeepAspectRatio,
                                                    Qt.TransformationMode.SmoothTransformation))

    def set_spacing(self, spacing):
        self.layout.setSpacing(spacing)

    def add_buddy(self, wdg: QWidget):
        index = self.layout.indexOf(self.text_label)
        if index < self.layout.count() - 1:
            self.layout.insertWidget(self.layout.count() - 1, wdg, stretch=1)
        else:
            self.layout.addWidget(wdg, stretch=1)
        self.layout.setStretch(index, 0)

    @staticmethod
    def create_label(text, icon_file=img.P_4_DOTS, buddy=None):
        ytl = YTitle(text, icon_path=icon_file, icon_size=QSize(12, 12), font_bold=False, font_size=9)
        if buddy is not None:
            ytl.add_buddy(buddy)
        return ytl


class YLink(Obj):
    def __init__(self, title, icon=None, target: Union[QWidget, Any] = None, tip_text=None, layout_flag='V', **kwargs):
        super().__init__(**kwargs)
        self.title = title
        self.icon = icon if icon else img.P_4_DOTS
        self.target = target
        self.tip_text = tip_text
        self.layout_flag = layout_flag

    def create_target(self):
        widget = create_widget(self.target) if self.target else QLabel(self.title)
        if widget and self.icon:
            widget.setWindowIcon(load_icon(self.icon))
        return widget

    @staticmethod
    def instance(p):
        if isinstance(p, YLink):
            return p
        elif isinstance(p, str):
            return YLink(p)
        elif isinstance(p, dict):
            return YLink(**p)
        return None


class YTab(QWidget):
    CloseTab = Signal(QObject)
    SwitchTab = Signal(QObject)

    def __init__(self, parent, ref_widget, title='', icon=None, close_able=False, tip_text=None, object_name=None):
        super().__init__(parent=parent)
        self.title = title
        self.ref_widget = ref_widget
        self.icon = icon if icon else img.P_4_DOTS
        self.setObjectName(object_name)
        self.tip_text = tip_text
        self.close_able = close_able
        self.ui = QObject(parent=self)
        self._is_active = False
        self.setup_ui()

    def setup_ui(self):
        vbox = QVBoxLayout(self)
        vbox.setSpacing(1)
        self.ui.btn_title = create_button(text=self.title,
                                          trigger=self.switch_tab,
                                          tip_text=self.tip_text,
                                          flat=True,
                                          icon_file=self.icon)
        self.ui.btn_title.setCheckable(True)

        hbox = create_layout(layout_flag='H')
        hbox.addWidget(self.ui.btn_title)
        if self.close_able:
            self.ui.btn_close = create_button(text='',
                                              trigger=self.close_tab,
                                              icon_file=img.P_CLOSE,
                                              flat=True)
            hbox.addWidget(self.ui.btn_close)
            self.ui.btn_close.setVisible(False)

        focus_bar = QFrame(parent=self)
        focus_bar.setFixedHeight(2)
        self.ui.focus_bar = focus_bar
        vbox.addLayout(hbox)
        vbox.addWidget(self.ui.focus_bar)
        self.hi_light(False)

    def close_tab(self):
        self.CloseTab.emit(self.ref_widget)
        self.close()

    def switch_tab(self):
        self.SwitchTab.emit(self.ref_widget)

    def activate(self, is_activate):
        self._is_active = is_activate
        self.ui.btn_title.setChecked(self._is_active)
        self.hi_light(self._is_active)
        change_font(self.ui.btn_title, bold=self._is_active)

    def hi_light(self, light):
        if self.is_active():
            active_color = QPalette().color(QPalette.Dark).name()
            self.ui.focus_bar.setFixedHeight(2)
            self.ui.btn_title.setFlat(False)
            self.ui.btn_title.setStyleSheet(f"background-color:{active_color}")
        else:
            self.ui.focus_bar.setFixedHeight(1)
            self.ui.btn_title.setFlat(False)
        if light or self.is_active():
            focus_color = QPalette().color(QPalette.Highlight).name()
            self.ui.focus_bar.setStyleSheet(f"background-color:{focus_color}")
        else:
            self.ui.focus_bar.setStyleSheet("")
            self.ui.btn_title.setFlat(True)
            self.ui.btn_title.setStyleSheet("")
        if self.close_able:
            self.ui.btn_close.setVisible(light)

    def enterEvent(self, event: QEnterEvent) -> None:
        super().enterEvent(event)
        self.hi_light(True)

    def leaveEvent(self, event: QEvent) -> None:
        super().leaveEvent(event)
        self.hi_light(False)

    def is_active(self):
        return self._is_active


class YWidgetPages(QWidget):
    """
    用StackedWidget和ButtonGroup组合的多页面组件.
    """
    pageSwitched = Signal(QWidget, YTab)
    pageClosed = Signal(QWidget, YTab)

    def __init__(self,
                 *pages,
                 parent=None,
                 close_able=False,
                 show_single_tab=False,
                 tab_alignment=Qt.AlignmentFlag.AlignLeft,
                 object_name=None,
                 ):
        """
        初始化方法
        :param pages: 页签定义信息
        :param parent:
        :param on_switch:
        :param show_single_tab:
        :param tab_alignment:
        """
        super().__init__(parent=parent)
        self._pages = [YLink.instance(p) for p in pages]  # 页签定义,支持多种参数类型
        self._tabs = []  # 页签控件
        self.close_able = close_able
        self.show_single_tab = show_single_tab
        self.tab_alignment = tab_alignment
        self.setObjectName(object_name)

        self._stacked_widget = None  # 堆叠widget控件
        self._tab_bar = None  # Tab栏控件

        self.setup_ui()

    def setup_ui(self):
        self.setLayout(create_layout('V'))

        # 页签按钮栏
        self._tab_bar = self.create_tab_bar()  # Tab栏控件
        self.layout().addWidget(self._tab_bar)
        self._tab_bar.layout().setAlignment(self.tab_alignment | Qt.AlignmentFlag.AlignVCenter)

        # 页签窗体
        self._stacked_widget = QStackedWidget(parent=self.parent())  # 堆叠widget控件
        self._stacked_widget.layout().setContentsMargins(0, 0, 0, 0)
        self._stacked_widget.layout().setSpacing(0)
        self.layout().addWidget(self._stacked_widget)

        # 创建各个页签以及切换按钮
        for pd in self._pages:
            self.add_page(pd)

        # 切换到第一个
        if self.page_count() > 0:
            self._tabs[0].activate(True)
            self._tabs[0].hi_light(False)

    def create_tab_bar(self):
        w = create_layout_widget(QWidget, parent=self, layout_flag='H')
        w.layout().setAlignment(Qt.AlignmentFlag.AlignVCenter | Qt.AlignmentFlag.AlignCenter)
        w.layout().setSpacing(12)
        return w

    def close_page(self, page):
        for tab in self._tabs:
            tab.activate(False)
        if page is None:
            return
        cur_index = self._stacked_widget.indexOf(page)
        self._stacked_widget.removeWidget(page)
        page.close()
        closed_tab = self._tabs.pop(cur_index)
        cur_index = self._stacked_widget.currentIndex()
        if cur_index >= 0:
            self.switch_page(cur_index)
            tab = self._tabs[cur_index]
            tab.activate(True)
            tab.hi_light(False)
        self.pageClosed.emit(page, closed_tab)

    def switch_page(self, page):
        for tab in self._tabs:
            tab.activate(False)
            tab.hi_light(False)
        if isinstance(page, QWidget):
            self._stacked_widget.setCurrentWidget(page)
        else:
            self._stacked_widget.setCurrentIndex(page)
        cur_tab = self._tabs[self._stacked_widget.currentIndex()]
        cur_tab.activate(True)
        self.pageSwitched.emit(self._stacked_widget.currentWidget(), cur_tab)

    def page(self, index):
        return self._pages[index]

    def pages(self):
        return self._pages

    def widget(self, index):
        return self._stacked_widget.widget(index)

    def current_widget(self) -> QWidget:
        return self._stacked_widget.currentWidget()

    def current_index(self) -> int:
        return self._stacked_widget.currentIndex()

    def page_count(self):
        return self._stacked_widget.count()

    def add_page(self, pd, auto_switch=False) -> QWidget:
        pd = YLink.instance(pd)
        cur_index = self._stacked_widget.currentIndex()
        if cur_index < 0:
            cur_index = 0

        # 创建页签主体Widget
        w = pd.create_target()
        if w.layout() is None:
            w.setLayout(create_layout(layout_flag=pd.layout_flag))
        w.layout().setAlignment(Qt.AlignmentFlag.AlignTop)

        # 创建Tab
        tab = self.create_tab(pd, w)

        self._stacked_widget.addWidget(w)
        self._tabs.append(tab)
        self._tab_bar.layout().addWidget(tab)

        self.show_tab_bar()
        if auto_switch:
            self.switch_page(w)
        else:
            self.switch_page(cur_index)  # 不要影响当前页
        return w

    def create_tab(self, pd, page):
        tab = YTab(parent=self,
                   ref_widget=page,
                   title=pd.title,
                   icon=pd.icon,
                   tip_text=pd.tip_text,
                   close_able=self.close_able,
                   )
        tab.SwitchTab.connect(self.switch_page)
        tab.CloseTab.connect(self.close_page)
        return tab

    def show_tab_bar(self):
        self._tab_bar.setVisible(len(self._tabs) > 1 or self.show_single_tab)


class WidgetDragger(QWidget):
    doubleClicked = Signal()

    def __init__(self, window, parent=None):
        QWidget.__init__(self, parent)
        self._windowPos = None
        self._mousePos = None
        self._window = window
        self._mousePressed = False

    def mousePressEvent(self, event):
        self._mousePressed = True
        self._mousePos = event.globalPos()
        self._windowPos = self._window.pos()

    def mouseMoveEvent(self, event):
        if self._mousePressed:
            self._window.move(self._windowPos +
                              (event.globalPos() - self._mousePos))

    def mouseReleaseEvent(self, event):
        self._mousePressed = False

    def mouseDoubleClickEvent(self, event):
        if self._window.windowState() == Qt.WindowState.WindowMaximized:
            self._window.showNormal()
        else:
            self._window.showMaximized()
        self._window.resizeEvent(None)
        self.doubleClicked.emit()


class YWindow(QWidget):
    TITLE_HEIGHT = 22

    def __init__(self, parent=None, full_left=True, object_name=None):
        super().__init__(parent=parent)
        self.ui = QObject(parent=self)
        self.full_left = full_left
        self.setObjectName(object_name)
        self.setup_ui()

    def setup_ui(self):
        self.setWindowFlags(
            Qt.WindowType.Window | Qt.WindowType.FramelessWindowHint | Qt.WindowType.WindowSystemMenuHint | Qt.WindowType.WindowMinimizeButtonHint
            | Qt.WindowType.WindowMaximizeButtonHint)
        self.create_widgets()
        if self.full_left:
            self.full_left_layout()
        else:
            self.v_tile_layout()
        self.refresh_size_bar()

    def create_widgets(self):
        self.ui.top = WidgetDragger(self, parent=self)
        top_layout: QHBoxLayout = create_layout(layout_flag='H')
        top_layout.setSpacing(1)
        self.ui.top.setLayout(top_layout)
        self.ui.title_bar = QWidget(parent=self)
        self.ui.size_bar = self.create_size_bar()
        top_layout.addWidget(self.ui.title_bar, 1)
        top_layout.addWidget(self.ui.size_bar, alignment=Qt.AlignmentFlag.AlignRight)

        self.ui.header = QWidget(parent=self)
        self.ui.header.setSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.ui.header.setVisible(False)
        self.ui.left = QWidget(parent=self)
        self.ui.left.setSizePolicy(QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        self.ui.left.setVisible(False)
        self.ui.content = QWidget(parent=self)
        self.ui.content.setSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        self.ui.bottom = QWidget(parent=self)
        self.ui.bottom.setSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.ui.bottom.setVisible(False)

    def full_left_layout(self):
        grid: QGridLayout = create_layout(layout_flag='G2')
        grid.setSpacing(0)
        # +----+------------------------------+
        # |    |  top (title_bar | size_bar)  |
        # |  l +------------------------------+
        # |  e |  header                      |
        # |  f +------------------------------+
        # |  t |  content                     |
        # -----+------------------------------+
        # |        b o t t o m                |
        # +-----------------------------------+
        grid.addWidget(self.ui.left, 0, 0, 3, 1)
        grid.addWidget(self.ui.top, 0, 1, 1, 1)
        grid.addWidget(self.ui.header, 1, 1, 1, 1)
        grid.addWidget(self.ui.content, 2, 1, 1, 1)
        grid.addWidget(self.ui.bottom, 3, 0, 1, 2)
        self.setLayout(grid)

    def v_tile_layout(self):
        """
        垂直平铺布局
        :return:
        """
        grid: QGridLayout = create_layout(layout_flag='G2')
        grid.setSpacing(0)
        # +----+------------------------------+
        # |     top (title_bar | size_bar)    |
        # +-----------------------------------+
        # |     header                        |
        # +------+----------------------------+
        # | left |     c o n t e n t          |
        # +------+----------------------------+
        # |        b o t t o m                |
        # +-----------------------------------+
        grid.addWidget(self.ui.top, 0, 0, 1, 2)
        grid.addWidget(self.ui.header, 1, 0, 1, 2)
        grid.addWidget(self.ui.left, 2, 0, 1, 1)
        grid.addWidget(self.ui.content, 2, 1, 1, 1)
        grid.addWidget(self.ui.bottom, 3, 0, 1, 2)
        self.setLayout(grid)

    def closeEvent(self, event):
        if pop_question('您确定要退出吗?'):
            event.accept()
        else:
            event.ignore()

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self.refresh_size_bar()

    def create_size_bar(self):
        buttons = {
            'min': create_tool_button(trigger=self.showMinimized,
                                      icon=load_icon(img.P_MIN, colored=False),
                                      tip_text='最小化',
                                      ),
            'max': create_tool_button(trigger=self.show_max_or_normal,
                                      icon=load_icon(img.P_MAX, colored=False),
                                      ),
            'close': create_tool_button(trigger=self.close,
                                        icon=load_icon(img.P_CLOSE, colored=False),
                                        tip_text='关闭',
                                        ),
        }
        bar = create_toolbar(buttons=buttons.values(), icon_size=self.TITLE_HEIGHT)
        bar.layout().setContentsMargins(0, 0, 0, 0)
        copy_obj(bar, buttons)
        return bar

    def refresh_size_bar(self):
        btn: QToolButton = getattr(self.ui.size_bar, 'max')
        is_max = self.windowState() == Qt.WindowState.WindowMaximized
        if is_max:
            btn.trigger = self.showNormal
            btn.setIcon(load_icon(img.P_SHOW_NORMAL, colored=False))
            btn.setToolTip('还原')
        else:
            btn.trigger = self.showMaximized
            btn.setIcon(load_icon(img.P_MAX, colored=False))
            btn.setToolTip('最大化')

    def show_max_or_normal(self):
        btn: QToolButton = getattr(self.ui.size_bar, 'max')
        is_max = get_attr(btn, '_is_max', self.windowState() == Qt.WindowState.WindowMaximized)
        if is_max:
            self.showNormal()
        else:
            self.showMaximized()
        set_attr(btn, '_is_max', not is_max)

    def title_bar(self) -> QWidget:
        return self.ui.title_bar

    def content(self) -> QWidget:
        return self.ui.content

    def left_widget(self) -> QWidget:
        return self.ui.left

    def header_widget(self) -> QWidget:
        return self.ui.header

    def bottom_widget(self) -> QWidget:
        return self.ui.bottom


class YNavPanel(QWidget):
    toLink = Signal(YLink)

    def __init__(self, parent=None, icon_size=20, icon_spacing=20, show_group=True, object_name=None):
        super().__init__(parent=parent)
        self.ui = QObject(parent=self)
        self.icon_size = icon_size
        self.icon_spacing = icon_spacing
        self.show_group = show_group
        self._links = {}
        self.setObjectName(object_name)
        self.setup_ui()

    def setup_ui(self):
        self.ui.group_menu = QMenu()
        self.ui.tabs = QTabWidget(parent=self)
        self.ui.tabs.setStyleSheet("pane{border:none;}")
        self.ui.tabs.tabBar().setVisible(False)
        self.ui.tabs.setToolTipDuration(0)
        self.setLayout(create_layout(layout_flag='V'))
        self.layout().addWidget(self.ui.tabs)

    def _add_group(self, group, tab_index):
        group.target = tab_index
        self.ui.group_menu.addAction(QAction(text=group.title, parent=self,
                                             triggered=lambda: self.switch_group(tab_index),
                                             icon=load_icon(group.icon),
                                             statusTip=group.tip_text))

    def add_links(self, group: YLink, items: list[YLink], logo=None):
        widget = create_layout_widget(layout_flag='V')
        widget.layout().setContentsMargins(0, 0, 0, 0)
        # widget.setAutoFillBackground(True)
        self._links[group] = items
        # 导航组切换按钮
        group_btn = create_menu_button(self.ui.group_menu,
                                       title=group.title,
                                       icon_file=group.icon,
                                       hovered=True,
                                       # tool_button=True,
                                       ) if self.show_group else None

        # 导航项->QAction
        actions = []
        for link in items:
            action = QAction(text=link.title,
                             triggered=self.to_link,
                             parent=self,
                             icon=load_icon(link.icon),
                             )
            action.setToolTip(link.tip_text)
            setattr(action, '_link', link)
            actions.append(action)

        # 导航工具条
        tbar = create_toolbar(actions=actions)
        tbar.setOrientation(Qt.Orientation.Vertical)
        tbar.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextUnderIcon)
        tbar.setIconSize(QSize(self.icon_size, self.icon_size))
        tbar.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        tbar.setStyleSheet(f'spacing:{self.icon_spacing}px;')
        tbar.setToolTip(group.tip_text)

        # 添加到widget
        if logo:
            widget.layout().addWidget(logo)
        if group_btn:
            widget.layout().addWidget(group_btn)
            widget.layout().addItem(
                QSpacerItem(2, self.icon_spacing, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum))
        widget.layout().addWidget(tbar)

        # 添加到tab中
        index = self.ui.tabs.addTab(widget, load_icon(group.icon), group.title)
        # 添加导航组
        self._add_group(group, index)

    def to_link(self):
        link = getattr(self.sender(), '_link')
        self.toLink.emit(link)

    def switch_group(self, index):
        grp_lst = list(self._links.keys())
        if not grp_lst:
            return
        group = grp_lst[index]
        tab_index = group.target
        self.ui.tabs.setCurrentIndex(tab_index)

    def showEvent(self, event):
        super().showEvent(event)
        self.switch_group(0)


class YLogo(QWidget):
    def __init__(self, img_file, logo_width=32, logo_height=32, parent=None, on_click=None, object_name=None):
        super().__init__(parent)
        self.setObjectName(object_name)
        vbox = QVBoxLayout(self)
        btn = QPushButton(parent=self)
        btn.setIcon(load_icon(img_file, colored=False))
        btn.setIconSize(QSize(logo_width, logo_height))
        btn.setFlat(True)
        if on_click:
            btn.clicked.connect(on_click)
        vbox.addWidget(btn)


class YearMonthSpin(QWidget):
    valueChanged = Signal(str, int, int)
    yearChanged = Signal(int)
    monthChanged = Signal(int)

    def __init__(self, cur_year=None, cur_month=None, parent=None, object_name=None):
        super().__init__(parent=parent)
        self.setObjectName(object_name)
        if cur_year is None:
            cur_year = datetime.today().year
        if cur_month is None:
            cur_month = datetime.today().month
        self.setLayout(QHBoxLayout(self))
        self.create_spnboxs()
        self._spn_year.setValue(cur_year)
        self._spn_month.setValue(cur_month)
        self.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.setFixedHeight(24)
        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        self.layout().setContentsMargins(0, 0, 0, 0)

    def create_spnboxs(self):
        self._spn_year = QSpinBox(self)
        self._spn_month = QSpinBox(self)
        self._spn_year.setRange(2000, 2099)
        self._spn_year.setSingleStep(1)
        self._spn_year.setSuffix(' 年 ')
        self._spn_month.setRange(1, 12)
        self._spn_month.setSuffix(' 月 ')
        self._spn_month.setSingleStep(1)
        self._spn_month.setWrapping(True)
        # self._spn_month.setButtonSymbols(QAbstractSpinBox.NoButtons)
        self.layout().addWidget(self._spn_year)
        self.layout().addWidget(self._spn_month)
        self._spn_year.valueChanged.connect(self.onYearChanged)
        self._spn_month.valueChanged.connect(self.onMonthChanged)

    def enclose_box(self):
        hbox = QHBoxLayout()
        hbox.addWidget(self)
        hbox.setAlignment(Qt.AlignmentFlag.AlignCenter)
        return hbox

    def month_code(self):
        return '%4d%02d' % (self.value())

    def value(self):
        return self.year(), self.month()

    def year(self):
        return self._spn_year.value()

    def month(self):
        return self._spn_month.value()

    def set_monthcode(self, mth_code):
        if mth_code:
            self.set_value(int(mth_code[:4]), int(mth_code[-2:]))

    def set_value(self, y, m):
        self._spn_year.setValue(y)
        self._spn_month.setValue(m)

    def start_date(self):
        y = self.year()
        m = self.month()
        return to_date('%4d-%02d-01' % (y, m))

    def end_date(self):
        y = self.year()
        m = self.month()
        if m >= 12:
            y += 1
            m = 1
        else:
            m += 1
        d = to_date('%4d-%02d-01' % (y, m))
        return d - timedelta(days=1)

    @Slot()
    def onYearChanged(self):
        self.yearChanged.emit(self.year())
        self.valueChanged.emit(self.month_code(), self.year(), self.month())

    @Slot()
    def onMonthChanged(self):
        self.monthChanged.emit(self.month())
        self.valueChanged.emit(self.month_code(), self.year(), self.month())


class YearSpin(QWidget):
    def __init__(self, parent=None, label='年度', value=None, on_change=None, label_left=True, object_name=None):
        super().__init__(parent=parent)
        self.setObjectName(object_name)
        self._label_left = label_left
        self._label = QLabel(label)
        self._spn_year = QSpinBox()
        self._spn_year.setRange(1900, 2999)
        self.set_on_change(on_change)
        if value is None:
            self._spn_year.setValue(datetime.today().year)
        else:
            self._spn_year.setValue(value)

        self.setup_ui()

    def set_on_change(self, on_change):
        if on_change:
            self._spn_year.valueChanged.connect(on_change)

    def setup_ui(self):
        hbox = QHBoxLayout(self)
        hbox.setContentsMargins(0, 0, 0, 0)
        if self._label_left:
            hbox.addWidget(self._label)
            hbox.addSpacing(2)
            hbox.addWidget(self._spn_year)
        else:
            hbox.addWidget(self._spn_year)
            hbox.addSpacing(2)
            hbox.addWidget(self._label)
        hbox.addStretch()

    def value(self):
        return self._spn_year.value()


class YearSelector(QWidget):
    valueChanged = Signal(int)

    def __init__(self, parent=None, value=None, on_change=None, object_name=None):
        super().__init__(parent)
        self.setObjectName(object_name)
        if value is None:
            value = datetime.now().year
        self._value = value

        self._label = QLabel()
        self._label.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)
        self._label.setAlignment(Qt.AlignmentFlag.AlignCenter | Qt.AlignmentFlag.AlignVCenter)
        change_font(self._label, bold=True, size=12)

        btn_pre = create_button('', trigger=lambda: self.changeValue(self._value - 1), icon_file=img.P_LEFT)
        btn_next = create_button('', trigger=lambda: self.changeValue(self._value + 1), icon_file=img.P_RIGHT)
        btn_pre.setFlat(True)
        btn_next.setFlat(True)
        btn_pre.setIconSize(QSize(16, 16))
        btn_next.setIconSize(QSize(16, 16))

        layout = QHBoxLayout(self)
        layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        layout.setSpacing(10)

        layout.addWidget(btn_pre)
        layout.addWidget(self._label)
        layout.addWidget(btn_next)

        self.setValue(value)
        self.set_on_change(on_change)

    def value(self):
        return self._value

    def changeValue(self, v):
        self.setValue(v)
        self.valueChanged.emit(v)

    def setValue(self, y):
        self._value = y
        self._label.setText('%4d 年' % self._value)

    def set_on_change(self, on_change):
        if on_change:
            self.valueChanged.connect(on_change)


class PeriodSelector(QWidget):
    valueChanged = Signal(object)

    def __init__(self, parent=None, cur_year=None, cur_month=None, ui_mode=0, show_switcher=True, object_name=None):
        super().__init__(parent)
        self.setObjectName(object_name)
        # False-by 年月； True- by 日期段
        self._ui_mode = ui_mode
        self._start_date = self._end_date = datetime.today()
        if cur_year is None:
            cur_year = datetime.today().year
        if cur_month is None:
            cur_month = datetime.today().month
        self._cur_year = cur_year
        self._cur_month = cur_month
        self._set_value_by_period()

        self._w_by_date = self._widget_by_date()
        self._w_by_ym = self._widget_by_ym()

        self._menu = QMenu(self)
        self.tab = QStackedWidget(self)
        self.tab.setMaximumHeight(60)
        self.tab.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        self.tab.addWidget(self._w_by_ym)
        self.tab.addWidget(self._w_by_date)
        self.tab.setCurrentIndex(self._ui_mode)

        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        layout.addWidget(self.tab)
        if show_switcher:
            layout.addSpacing(80)
            layout.addWidget(self.create_mode_switcher())

    # 两个日期选择时间段
    def _widget_by_date(self):
        w = QWidget()
        self.edt_sdate = create_date_edit(self._start_date)
        self.edt_edate = create_date_edit(self._end_date)
        change_font(self.edt_sdate, size=12)
        change_font(self.edt_edate, size=12)

        layout = QHBoxLayout(w)
        layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        layout.setSpacing(10)

        layout.addWidget(QLabel('从'))
        layout.addWidget(self.edt_sdate)
        layout.addWidget(QLabel('至'))
        layout.addWidget(self.edt_edate)

        btn_apply = create_button(trigger=self._apply_date_edit,
                                  icon_file=img.P_REFRESH,
                                  flat=True,
                                  tip_text='确定日期范围')
        layout.addWidget(btn_apply)

        return w

    # 年/月方式
    def _widget_by_ym(self):
        w = QWidget()
        self._label = QLabel()
        self._label.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)
        self._label.setAlignment(Qt.AlignmentFlag.AlignCenter | Qt.AlignmentFlag.AlignVCenter)

        btn_rev_fst = create_button(trigger=lambda: self._apply_ym(y=-1),
                                    icon_file=img.P_LEFT2,
                                    flat=True,
                                    tip_text='前一年')
        btn_rev = create_button(trigger=lambda: self._apply_ym(m=-1),
                                icon_file=img.P_LEFT,
                                flat=True,
                                tip_text='前一月')
        btn_fwd = create_button(trigger=lambda: self._apply_ym(m=1),
                                icon_file=img.P_RIGHT,
                                flat=True,
                                tip_text='后一月')
        btn_fwd_fst = create_button(trigger=lambda: self._apply_ym(y=1),
                                    icon_file=img.P_RIGHT2,
                                    flat=True,
                                    tip_text='后一年')

        btn_fwd.setIconSize(QSize(16, 16))
        btn_fwd_fst.setIconSize(QSize(16, 16))

        layout = QHBoxLayout(w)
        layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        layout.setSpacing(10)

        layout.addWidget(btn_rev_fst)
        layout.addWidget(btn_rev)
        layout.addWidget(self._label)
        layout.addWidget(btn_fwd)
        layout.addWidget(btn_fwd_fst)

        btn_apply = create_button(trigger=self._apply_ym,
                                  icon_file=img.P_REFRESH,
                                  flat=True,
                                  tip_text='刷新')
        layout.setSpacing(40)
        layout.addWidget(btn_apply)

        self._refresh()
        return w

    def create_mode_switcher(self):
        self._menu.addActions([
            yui.create_action('年/月选择', icon_file=img.P_INDEX, trigger=lambda: self.switch_to(0), parent=self, ),
            yui.create_action('日期段选择', icon_file=img.P_CALENDAR, trigger=lambda: self.switch_to(1), parent=self, ),
        ])
        self._menu.addSeparator()
        self._menu.addActions([
            yui.create_action('自年初起', icon_file=img.P_LOOP, trigger=lambda: self.switch_to(2), parent=self),
            yui.create_action('上年同期', icon_file=img.P_LAST, trigger=lambda: self.switch_to(3), parent=self),
            yui.create_action('下年同期', icon_file=img.P_NEXT, trigger=lambda: self.switch_to(4), parent=self),
        ])

        btn = create_menu_button(self._menu, icon_file=img.P_3DOTS_V, hovered=True)
        return btn

    def switch_to(self, mode):
        if mode in (0, 1):
            self._ui_mode = mode
            self.tab.setCurrentIndex(self._ui_mode)
            return
        self.tab.setCurrentIndex(1)
        if mode == 2:  # 自年初起
            self.edt_sdate.setDate(QDate(self._start_date.year, 1, 1))
        elif mode == 3:  # 上年同期
            self.edt_sdate.setDate(QDate(self._start_date.year - 1, self._start_date.month, self._start_date.day))
            self.edt_edate.setDate(QDate(self._end_date.year - 1, self._end_date.month, self._end_date.day))
        elif mode == 4:  # 下年同期
            self.edt_sdate.setDate(QDate(self._start_date.year + 1, self._start_date.month, self._start_date.day))
            self.edt_edate.setDate(QDate(self._end_date.year + 1, self._end_date.month, self._end_date.day))
        self._apply_date_edit()

    def set_date_pair(self, start_date, end_date):
        self.switch_to(1)  # 按照日期段方式设置，强制切换为日期段选择
        self._start_date, self._end_date = to_date(start_date), to_date(end_date)
        self.edt_sdate.setDate(QDate(self._start_date.year, self._start_date.month, self._start_date.day))
        self.edt_edate.setDate(QDate(self._end_date.year, self._end_date.month, self._end_date.day))

    def value(self) -> (object, object):
        return self._start_date, self._end_date

    def str_value(self) -> (str, str):
        return format_date(self._start_date), format_date(self._end_date)

    def _apply_date_edit(self):
        self._start_date = self.edt_sdate.date().toPython()
        self._end_date = self.edt_edate.date().toPython()
        self.valueChanged.emit(self)

    def _apply_ym(self, y=0, m=0):
        ny = self._cur_year + y
        nm = self._cur_month + m
        if nm > 12:
            nm = 1
            ny += 1
        if nm < 1:
            nm = 12
            ny -= 1
        self._cur_year = ny
        self._cur_month = nm
        self._set_value_by_period()
        self._refresh()
        self.valueChanged.emit(self)

    def _refresh(self):
        change_font(self._label, size=12, bold=True)
        self._label.setText(' %4d 年 %2d 月' % (self._cur_year, self._cur_month))

    def _set_value_by_period(self):
        self._start_date = to_date('%4d-%02d-01' % (self._cur_year, self._cur_month))
        y = self._cur_year
        m = self._cur_month
        if m >= 12:
            y += 1
            m = 1
        else:
            m += 1
        d = to_date('%4d-%02d-01' % (y, m))
        self._end_date = d - timedelta(days=1)

    def month_code(self):
        return '%4d%2d' % (self._cur_year, self._cur_month)

    def set_month_code(self, mth_code):
        if mth_code:
            (self._cur_year, self._cur_month) = (int(mth_code[:4]), int(mth_code[-2:]))
            self._set_value_by_period()
        self._refresh()

    def year(self):
        return self._cur_year

    def month(self):
        return self._cur_month

    def start_date(self):
        return self._start_date

    def end_date(self):
        return self._end_date

    def enclose_box(self):
        hbox = QHBoxLayout()
        hbox.addWidget(self)
        hbox.setAlignment(Qt.AlignmentFlag.AlignCenter)
        return hbox


class TabInfo:
    def __init__(self, title, icon=None, w_class=None, layout_flag=None):
        self.title = title
        self.icon = icon if icon else img.P_4_DOTS
        self.w_class = w_class if w_class else QWidget
        self.layout_flag = layout_flag if layout_flag else 'V'


class TabSetting:
    """
    描述多页签信息
    """

    def __init__(self, *args, stretch=True):
        """
        多页签信息。
        :param stretch: 切换按钮栏是否在右部留白。False - 不留白，按钮会均匀分布占满顶部栏位; True - 右部留白
        :param args: 可以是字符串列表或字典列表。
        如果是字典或TabInfo的列表，每个元素对应一个页签的描述，分别为:
            'title':    标题
            'icon':     图标文件名，默认为 'bullet3.png'
            'w_class':  页Widget类，默认为QWidget
            'layout_flag': 页布局标记，默认为'V'
        如果是字符串列表，创建的页签使用列表元素作为标题，其它属性为默认值。
        """
        self._data = []
        self.stretch = stretch
        for arg in args:
            if is_collection(arg):
                self._data.extend(arg)
            else:
                if type(arg) is str:  # 只使用标题
                    self._data.append(TabInfo(arg))
                else:
                    r = TabInfo('')
                    copy_obj(r, arg)
                    self._data.append(r)

    def count(self):
        return len(self._data)

    def items(self):
        return self._data

    def index_of(self, index):
        tab_index = 0
        if isinstance(index, int):
            if index < 0 or index > len(self._data) - 1:
                raise IndexError(f'子页面数量：{len(self._data)}', index)
            tab_index = index
        else:
            str_index = str(index).strip()
            for index, text in enumerate([t.title for t in self._data]):
                if str_index == text.strip():
                    tab_index = index
        return tab_index

    def item(self, index):
        return self._data[self.index_of(index)]


class YStackedPages(QWidget):
    WIDGET_TAG = '_widget'
    LAYOUT_TAG = '_m_layout'

    """
    用StackedWidget和ButtonGroup组合的多页面组件.
    """

    def __init__(self, page_def: TabSetting = None, parent=None, on_switch=None, object_name=None):
        """
        初始化方法
        :param page_def: 页签定义信息
        :param parent:
        :param on_switch:
        """
        super().__init__(parent=parent)
        self._button_group = QButtonGroup(parent=self)
        self._pages = QStackedWidget(parent=self)
        self._tab_bar = self.create_tab_bar()
        self._on_switch = on_switch
        self._page_def = page_def
        self.setObjectName(object_name)

        self.setLayout(create_layout('V'))
        # 页签按钮栏
        self.layout().addWidget(self._tab_bar)
        # 页签窗体
        self.layout().addWidget(self._pages)

        # 创建各个页签以及切换按钮
        for pd in self._page_def.items():
            self.add_page(pd)

        if self._page_def.stretch:
            self._tab_bar.layout().addStretch(0)  # 右部留白

    def create_tab_bar(self):
        w = create_layout_widget(QWidget, parent=self, layout_flag='H')
        # w.layout().setAlignment(Qt.AlignmentFlag.AlignVCenter | Qt.AlignmentFlag.AlignCenter)
        w.layout().setSpacing(12)
        w.layout().setContentsMargins(0, 8, 0, 8)
        return w

    def switch_widget(self):
        btn = self.sender()
        widget = get_attr(btn, self.WIDGET_TAG)
        self._pages.setCurrentWidget(widget)
        if self._on_switch:
            self._on_switch(self._pages, widget)

    def page(self, idx):
        buttons = self._button_group.buttons()
        widget_index = self._page_def.index_of(idx)
        return get_attr(buttons[widget_index], self.WIDGET_TAG)

    def pages(self):
        return self._pages

    def page_layout(self, idx) -> QLayout:
        return get_attr(self.page(idx), self.LAYOUT_TAG)

    def add_page(self, pd):
        cur_index = self._pages.currentIndex()
        if cur_index < 0:
            cur_index = 0

        # 创建页签Widget
        w = create_layout_widget(pd.w_class, layout_flag='V')
        layout = create_layout(layout_flag=pd.layout_flag)
        set_attr(w, self.LAYOUT_TAG, layout)
        w.layout().addLayout(layout)
        w.layout().setAlignment(Qt.AlignmentFlag.AlignTop)
        self._pages.addWidget(w)

        # 创建切换按钮
        button = create_button(text=pd.title,
                               trigger=self.switch_widget,
                               flat=True,
                               icon_file=pd.icon)
        button.setCheckable(True)
        set_attr(button, self.WIDGET_TAG, w)  # 将表单对象保存到按钮对象的属性中供切换按钮时获取
        self._button_group.addButton(button)
        self._button_group.setExclusive(True)
        self._tab_bar.layout().addWidget(button)
        self._pages.setCurrentIndex(cur_index)  # 不要影响当前页
        self.show_buttons()

    def show_buttons(self):
        self._tab_bar.setVisible(len(self._button_group.buttons()) > 1)


class YPercentCircle(QWidget):
    def __init__(self, parent=None, object_name=None):
        super().__init__(parent)
        self.setObjectName(object_name)
        self.value = 0
        self.width = 100
        self.height = 100
        self.progress_width = 10
        self.progress_rounded_cap = True
        self.progress_color = QColor(0, 137, 255)
        self.max_value = 100
        self.font_family = "Arial"
        self.font_size = 16
        self.suffix = "%"
        self.text_color = QColor(192, 192, 192)
        self.enable_bg = True
        self.bg_color = QColor(220, 220, 220)
        """
        通过修改以下属性来自定义组件外观：
        progress.progress_width = 15  # 圆环宽度
        progress.progress_color = QColor(255, 0, 127)  # 粉红色进度条
        progress.bg_color = QColor(240, 240, 240)  # 浅灰色背景
        progress.text_color = QColor(70, 70, 70)  # 深灰色文字
        progress.font_size = 16  # 更大的字体
        progress.suffix = "％"  # 使用全角百分号
        """
        # 设置默认大小
        self.setMinimumSize(self.width, self.height)

    def set_value(self, value):
        self.value = value
        self.update()  # 触发重绘

    def paintEvent(self, event):
        # 设置长宽
        width = self.width - self.progress_width
        height = self.height - self.progress_width
        margin = self.progress_width / 2
        value = self.value * 360 / self.max_value

        # 绘制器
        paint = QPainter()
        paint.begin(self)
        paint.setRenderHint(QPainter.RenderHint.Antialiasing)

        # 创建矩形区域
        rect = QRectF(margin, margin, width, height)

        # 绘制背景圆环
        if self.enable_bg:
            pen = QPen()
            pen.setColor(self.bg_color)
            pen.setWidth(self.progress_width)
            if self.progress_rounded_cap:
                pen.setCapStyle(Qt.PenCapStyle.RoundCap)
            paint.setPen(pen)
            paint.drawArc(rect, 0, 360 * 16)

        # 绘制进度圆环
        pen = QPen()
        pen.setColor(self.progress_color)
        pen.setWidth(self.progress_width)
        if self.progress_rounded_cap:
            pen.setCapStyle(Qt.PenCapStyle.RoundCap)
        paint.setPen(pen)
        paint.drawArc(rect, 90 * 16, -value * 16)

        # 绘制文本
        font = QFont(self.font_family, self.font_size)
        font.setBold(True)
        paint.setFont(font)
        paint.setPen(QColor(self.text_color))
        paint.drawText(rect, Qt.AlignmentFlag.AlignCenter, f"{self.value}{self.suffix}")

        paint.end()


class YPercentBar(QWidget):
    def __init__(self, parent=None, value=0.0, orientation=Qt.Orientation.Vertical, object_name=None, **kwargs):
        super().__init__(parent=parent, **kwargs)
        self.value = value
        self.orientation = orientation
        self.setObjectName(object_name)
        self.setup_ui()
        self.set_orientation(self.orientation)
        self.set_value(self.value)

    def setup_ui(self):
        self.setLayout(create_layout(layout_flag='V'))
        pbar = QProgressBar(parent=self, minimum=0, maximum=100)
        pbar.setObjectName('pbar')
        pbar.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.layout().addWidget(pbar)

    def set_value(self, value):
        self.value = value
        pbar: QProgressBar = self.findChild(QProgressBar, 'pbar')
        pbar.setValue(self.value)

    def set_orientation(self, value):
        self.orientation = value
        pbar: QProgressBar = self.findChild(QProgressBar, 'pbar')
        pbar.setOrientation(self.orientation)


class YCard(QFrame):
    def __init__(self, title_text,
                 items: list[object],
                 parent=None,
                 grid_col_num=3,
                 title_icon=None,
                 title_alignment=Qt.AlignmentFlag.AlignCenter,
                 caption_alignment=Qt.AlignmentFlag.AlignLeft,
                 value_alignment=Qt.AlignmentFlag.AlignRight,
                 comment_alignment=Qt.AlignmentFlag.AlignLeft,
                 bg_color=None,
                 on_title_clicked=None,
                 tag=None,
                 object_name=None,
                 ):
        super().__init__(parent=parent)
        self.title_text = title_text
        self.title_icon = title_icon
        self.title_alignment = title_alignment
        self.caption_alignment = caption_alignment
        self.value_alignment = value_alignment
        self.comment_alignment = comment_alignment
        self.setObjectName(object_name)
        self.items = items
        self.on_title_clicked = on_title_clicked
        self.tag = tag
        self.grid_col_num = grid_col_num
        self.background_color = bg_color
        self.ui = Obj()
        self.setup_ui()

    @staticmethod
    def format_value(v) -> str:
        if v is None:
            return ''
        if isinstance(v, str):
            return v
        if isinstance(v, Decimal) or isinstance(v, float):
            return f"{to_float(v):,.2f}"
        return str(v)

    def setup_ui(self):
        self.setMinimumSize(QSize(100, 50))
        gbox = QGridLayout()
        gbox.setHorizontalSpacing(16)
        gbox.setVerticalSpacing(8)
        self.setLayout(gbox)
        self.ui.title = YTitle(self.title_text,
                               parent=self,
                               icon_path=self.title_icon,
                               alignment=self.title_alignment,
                               icon_size=QSize(20, 20),
                               font_size=12,
                               on_click=self.on_title_clicked,
                               click_param=self.tag,
                               )
        gbox.addWidget(self.ui.title, 0, 0, 1, self.grid_col_num)
        row = 1
        for i, item in enumerate(self.items):
            caption = get_attr(item, 'caption', '')
            value = get_attr(item, 'value', '')
            comment = get_attr(item, 'comment', '')

            lb_caption = QLabel(f"{caption}", alignment=self.caption_alignment)
            lb_value = QLabel(self.format_value(value), alignment=self.value_alignment)
            lb_comment = QLabel(f"{comment}", alignment=self.comment_alignment)

            color = get_attr(item, 'color')
            if color:
                set_label_color(lb_value, fore_color=color)
            color = get_attr(item, 'caption_color')
            if color:
                set_label_color(lb_caption, fore_color=color)
            color = get_attr(item, 'comment_color')
            if color:
                set_label_color(lb_comment, fore_color=color)

            font_size = get_attr(item, 'font_size')
            if font_size:
                change_font(lb_value, size=font_size)

            gbox.addWidget(lb_caption, row + i, 0)
            gbox.addWidget(lb_value, row + i, 1)
            gbox.addWidget(lb_comment, row + i, 2)
            set_attr(self.ui, f'lb_value_{i}', lb_value)
        self.set_background_color(self.background_color)

    def update_items(self, items):
        self.items = items
        for i, item in enumerate(self.items):
            value = get_attr(item, 'value', '')
            lb: QLabel = get_attr(self.ui, f'lb_value_{i}')
            if lb:
                lb.setText(self.format_value(value))

    def set_background_color(self, bg_color):
        self.background_color = bg_color
        if self.background_color:
            self.setStyleSheet(f"background-color:{self.background_color};border-radius: 12px;")


class YSlideBoard(QWidget):
    def __init__(self, parent=None, object_name=None):
        super().__init__(parent)
        self.setObjectName(object_name)
        self._widget_width = 200
        self._button_size = QSize(40, 100)
        self._spacing = 10

        self._init_ui()
        self._setup_connections()

    def _init_ui(self):
        self.main_layout = QHBoxLayout(self)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(0)

        # 左按钮
        self.left_button = QPushButton("◀")
        self.left_button.setFlat(True)
        self.left_button.setFixedSize(self._button_size)
        self.left_button.setVisible(False)

        # 滚动区域
        self.scroll_area = QScrollArea()
        self.scroll_area.setWidgetResizable(False)
        self.scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.scroll_area.setFrameShape(QFrame.Shape.NoFrame)

        # 容器widget
        self.container = QWidget()
        self.container_layout = QHBoxLayout(self.container)
        self.container_layout.setContentsMargins(0, 0, 0, 0)
        self.container_layout.setSpacing(self._spacing)

        self.scroll_area.setLayout(QHBoxLayout())
        self.scroll_area.setWidget(self.container)

        # 右按钮
        self.right_button = QPushButton("▶")
        self.right_button.setFlat(True)
        self.right_button.setFixedSize(self._button_size)
        self.right_button.setVisible(False)

        self.main_layout.addWidget(self.left_button)
        self.main_layout.addWidget(self.scroll_area)
        self.main_layout.addWidget(self.right_button)

    def _setup_connections(self):
        self.left_button.clicked.connect(self._slide_left)
        self.right_button.clicked.connect(self._slide_right)

    def set_button_size(self, size: QSize):
        self._button_size = size
        self.left_button.setFixedSize(size)
        self.right_button.setFixedSize(size)

    def set_widget_width(self, width: int):
        self._widget_width = width
        for i in range(self.container_layout.count()):
            widget = self.container_layout.itemAt(i).widget()
            if widget:
                widget.setFixedWidth(width)
        self._update_container_size()
        self._check_buttons_needed()  # 新增：立即检查按钮是否需要显示

    def set_spacing(self, spacing: int):
        self._spacing = spacing
        self.container_layout.setSpacing(spacing)
        self._update_container_size()
        self._check_buttons_needed()  # 新增：立即检查按钮是否需要显示

    def add_widget(self, widget: QWidget):
        widget.setFixedWidth(self._widget_width)
        widget.setMinimumHeight(150)
        self.container_layout.addWidget(widget)
        self._update_container_size()
        self._check_buttons_needed()  # 新增：立即检查按钮是否需要显示

    def _update_container_size(self):
        count = self.container_layout.count()
        total_width = count * (self._widget_width + self._spacing) - self._spacing
        self.container.setFixedSize(total_width, self.height())

    def _check_buttons_needed(self):
        """新增方法：检查是否需要显示滑动按钮"""
        if not self.isVisible():
            return

        # 强制更新布局，确保获取正确的尺寸
        self.updateGeometry()
        QApplication.processEvents()

        container_width = self.container.width()
        viewport_width = self.scroll_area.viewport().width()

        # 如果内容宽度大于可视区域宽度，则需要显示按钮
        needs_buttons = container_width > viewport_width

        self.right_button.setVisible(needs_buttons)
        self.left_button.setVisible(needs_buttons and
                                    self.scroll_area.horizontalScrollBar().value() > 0)

    def _slide_left(self):
        scroll_bar = self.scroll_area.horizontalScrollBar()
        current = scroll_bar.value()
        step = self._widget_width + self._spacing
        scroll_bar.setValue(max(0, current - step))
        self._update_button_visibility()

    def _slide_right(self):
        scroll_bar = self.scroll_area.horizontalScrollBar()
        current = scroll_bar.value()
        step = self._widget_width + self._spacing
        max_val = scroll_bar.maximum()
        scroll_bar.setValue(min(max_val, current + step))
        self._update_button_visibility()

    def _update_button_visibility(self):
        scroll_bar = self.scroll_area.horizontalScrollBar()
        current_scroll = scroll_bar.value()
        max_scroll = scroll_bar.maximum()

        self.left_button.setVisible(current_scroll > 0)
        self.right_button.setVisible(current_scroll < max_scroll)

    def showEvent(self, event):
        """重写showEvent以确保窗口显示时正确计算按钮可见性"""
        super().showEvent(event)
        self._check_buttons_needed()

    def resizeEvent(self, event):
        self._update_container_size()
        self._check_buttons_needed()
        super().resizeEvent(event)


class YSortableItem(QTableWidgetItem):
    DR_AttrName = 'DATA_ROLE_OF_SORT'

    def set_data_role_of_sort(self, data_role: Qt.ItemDataRole):
        set_attr(self, YSortableItem.DR_AttrName, data_role)

    def data_role_of_sort(self):
        return get_attr(self, YSortableItem.DR_AttrName, Qt.ItemDataRole.UserRole)

    def __lt__(self, other):
        # 比较原始数值,用于排序
        role = self.data_role_of_sort()
        d1, d2 = self.data(role), other.data(role)
        if d1 is None:
            return True
        if d2 is None:
            return False
        return d1 < d2


class YSortableTable(QTableWidget):
    def __init__(self, parent=None, columns_width=None):
        super().__init__(parent)
        self.setStyleSheet('gridline-color:palette(Midlight)')
        self.setMouseTracking(True)
        self.original_style = None
        self.columns_width = columns_width
        self.current_cell = (-1, -1)
        self.sort_orders = {}
        self.setSortingEnabled(True)  # 启用表头排序
        self.horizontalHeader().clicked.connect(self.header_clicked)
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectItems)  # 改为单元格选择

    def mouseMoveEvent(self, event):
        item = self.itemAt(event.position().toPoint())
        if item:
            row, col = item.row(), item.column()
            self.highlight_cell(row, col)
        super().mouseMoveEvent(event)

    def header_clicked(self, col_index):
        order = self.sort_orders.get(col_index, Qt.SortOrder.DescendingOrder)
        order = Qt.SortOrder.DescendingOrder if order == Qt.SortOrder.AscendingOrder else Qt.SortOrder.DescendingOrder
        self.sortByColumn(col_index, order)
        self.horizontalHeader().setSortIndicator(col_index, order)
        self.sort_orders[col_index] = order

    def highlight_cell(self, row, col):
        # 恢复之前高亮的单元格
        if self.current_cell != (-1, -1):
            prev_item = self.item(*self.current_cell)
            if prev_item and self.original_style:
                prev_item.setForeground(self.original_style)

        # 高亮当前单元格
        item = self.item(row, col)
        if item:
            self.original_style = item.foreground()
            item.setForeground(yui.current_theme.hover_color())
            self.current_cell = (row, col)

    def adjust_column_width(self):
        if self.columns_width:
            for i, w in enumerate(self.columns_width):
                self.setColumnWidth(i, w)
        else:
            self.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeMode.ResizeToContents)
            self.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeMode.Interactive)


def show_widget(target: QWidget | YLink):
    m_win = get_mainwindow()
    fn_show_widget = get_attr(m_win, "show_widget")
    if isinstance(target, YLink) and fn_show_widget is not None:
        fn_show_widget(target)
    elif isinstance(target, QWidget):
        if fn_show_widget is not None:
            fn_show_widget(YLink(target.windowTitle(), icon=img.P_DOC, target=target))
        else:
            set_attr(m_win, '_target_widget_', target)
            target.show()


def switch_to_widget(widget: QWidget):
    m_win = get_mainwindow()
    fn = get_attr(m_win, "switch_to_widget")
    if fn:
        fn(widget)


class Redirector(QObject):
    displayText = Signal(str)
    reset = Signal()

    def __init__(self, parent=None):
        super().__init__(parent)

    def write(self, text):
        self.displayText.emit(str(text))

    def writelines(self, lines):
        for line in lines:
            self.write(line)

    def start_fn(self, func, *args, **kwargs):
        """
        执行函数，并重定向输出
        :param func:
        :param args:
        :param kwargs:
        :return:
        """
        saved_streams = sys.stdin, sys.stdout  # 保存标准输入输出对象，以便恢复
        sys.stdout = self
        sys.stderr = sys.stdout
        self.reset.emit()
        result = func(*args, **kwargs)
        sys.stdin, sys.stdout = saved_streams
        return result



# -*- __init__.py: python ; coding: utf-8 -*-
import base64
import datetime
import json
import os
from decimal import Decimal

import pandas as pd

b64_char = r"0123456789KLMNOPQRSTABCDEFGHIJefghijklmnopqrstuvwxUVWXYZabcdyz+/="
m64_char = r"fehimgjlnk1025734896LKNOSMPRTQBADEICFHJGVUXYcWZbdaporqvstwxuzy.$^"


def has_attr(obj, attr):
    if obj is None:
        return False
    if isinstance(obj, dict):
        return obj.get(attr) is not None
    else:
        return hasattr(obj, attr)


def get_attr(obj, attr, no_attr=None):
    """
    返回对象的给定属性
    :param obj: 对象/dict
    :param attr: 属性名/key
    :param no_attr: 如果属性不存在，返回本值
    :return:
    """
    if obj is None:
        return no_attr
    if isinstance(obj, dict):
        return obj.get(attr, no_attr)
    else:
        return getattr(obj, attr) if hasattr(obj, attr) else no_attr


def set_attr(obj, attr, value):
    if obj is None:
        return
    if isinstance(obj, dict):
        obj[attr] = value
    else:
        setattr(obj, attr, value)


def set_default_values(o, **kwargs):
    """
    使用给出的名值对设置对象的属性值。
    若kwargs中的key不在o的原属性中，直接在o中添加属性及属性值，若已存在key属性，仅在原有属性值为None时才使用新值。
    :param o:
    :param kwargs:
    :return:
    """
    for attr in kwargs:
        if not hasattr(o, attr):
            setattr(o, attr, kwargs[attr])
        else:
            if getattr(o, attr) is None:
                setattr(o, attr, kwargs[attr])
    return o


def copy_obj(o_tar, o_src, skip_keys=None):
    """
    复制对象属性值到另外的对象中
    :param o_tar: 目标对象
    :param o_src: 源对象
    :param skip_keys: 忽略的属性(list)
    :return:
    """
    if skip_keys is None:
        skip_keys = []
    keys = o_src.keys() if isinstance(o_src, dict) else o_src.__dict__.keys()
    for key in keys:
        if key not in skip_keys and not key.startswith('__'):
            v = get_attr(o_src, key)
            set_attr(o_tar, key, v)
    return o_tar


def format_date(d):
    if not d or pd.isna(d):
        return None
    return ('%s' % d)[:10]


def format_datetime(d, fmt='%Y-%m-%d %H:%M:%S.%f'):
    if not d or pd.isna(d):
        return ''
    import datetime
    if d is datetime.date:
        return datetime.datetime.strftime(d, fmt)
    else:
        return d


def format_id(n, fmt='%06d'):
    if n is None or pd.isna(n):
        return ''
    else:
        return fmt % n


def format_percent(n, fmt='%.2f%%'):
    if n is None or pd.isna(n):
        return ''
    else:
        return fmt % to_float(n)


def format_year_mth(d):
    ret = ''
    s = format_date(d)
    if s and len(s) == 10:
        ret = s[2:4] + s[5:7]
    return ret


def format_obj(obj, exp, use_repr=False):
    """
    对象属性表达式做格式化转换，支持使用repr。
    例如 obj.name='ABC', obj.id=1000
    format_obj(obj,'{id} - {name}') 返回 '1000 - ABC'
    :param obj: 对象
    :param exp: 表达式，如'{attr}'
    :param use_repr: 替换变量时是否用repr
    :return:
    """
    if obj is None:
        return exp
    dv = obj if isinstance(obj, dict) else obj.__dict__
    if use_repr:
        dv = {k: repr(v) for k, v in dv.items()}
    f_exp = f"f\"{exp}\""
    return eval(f_exp, dv)


def to_float(o, none_value=0.0) -> float:
    if o is None or pd.isna(o):
        return none_value
    text = str(o).replace(',', '').replace('%', '').replace(' ', '')
    return float(text) if text else none_value


def to_int(o, none_value=0) -> int:
    if o is None or pd.isna(o):
        return none_value
    text = str(o).replace(',', '').replace('%', '').replace(' ', '')
    return int(text) if text else none_value


def to_date(d_str):
    return to_datetime(d_str, '%Y-%m-%d')


def to_timestamp(dt_str, fmt='%Y-%m-%d %H:%M:%S.%f'):
    return to_datetime(dt_str, fmt)


def to_datetime(d_str, fmt='%Y-%m-%d %H:%M:%S'):
    import datetime
    if not d_str or pd.isna(d_str):
        return None
    if hasattr(d_str, 'strip'):
        ds = d_str.strip()
        if len(ds) >= 10:
            ds = ds[:10]
        return datetime.datetime.strptime(ds, fmt)
    else:
        return d_str


def safe_sum(*val):
    total = 0.0
    if len(val) == 1:
        lst = val[0]
    else:
        lst = val
    for v in lst:
        total += to_float(v)
    return total


def safe_add(a, b):
    return to_float(a, none_value=0) + to_float(b, none_value=0)


def safe_diff(a, b):
    return to_float(a, none_value=0) - to_float(b, none_value=0)


def safe_mul(a, b, none_value=0):
    if a and b:
        return a * b
    else:
        return none_value


def safe_div(a, b, zero_div=None, none_as=None):
    if a is None or b is None:
        return none_as
    try:
        if isinstance(a, Decimal):
            a = to_float(a)
        if isinstance(b, Decimal):
            b = to_float(b)
        return a / b
    except ZeroDivisionError:
        return zero_div


def safe_percent(a, b, delta=False):
    if b is None:
        return None
    if abs(b) < 0.001:
        return None
    a = to_float(a)
    b = to_float(b)
    if delta:
        return (a / b - 1) * 100
    else:
        return a / b * 100


def is_execute_able(f):
    """
    判断一个名称是否为可执行类型即：FunctionType,MethodType,LambdaType
    :param f:
    :return:
    """
    import types
    return isinstance(f, types.FunctionType) or isinstance(f, types.MethodType) or isinstance(f, types.LambdaType)


def is_collection(o):
    """
    判断对象是否为集合类型，包括list,tuple,set和dict
    :param o:
    :return:
    """
    return isinstance(o, (list, tuple, set, dict))


def create_instance(mod_name, cls_name, *args, **kwargs):
    """
    :param mod_name: 模块名称
    :param cls_name: 类名称或函数名称
    :param args: 实例化或函数调用的参数表
    :param kwargs: 实例化或函数调用的参数表
    :return:
    """
    mod = __import__(mod_name, globals(), locals(), [cls_name])
    proc = getattr(mod, cls_name)
    if is_execute_able(proc):  # 是函数
        return proc(*args, **kwargs)
    else:  # 实例化类
        return proc(*args, **kwargs)


def call_func(mod_name, func_name, *args, **kwargs):
    """
    根据模块名称和函数/类名称调用函数或创建对象。
    :param mod_name: 模块名称，如 'yut.gui'
    :param func_name: 函数或类名称，如 'create_button' 或 'HoveredButton'
    :param args:
    :param kwargs:
    :return:
    """
    mod = __import__(mod_name, globals(), locals(), [func_name])
    fn = getattr(mod, func_name)
    return fn(*args, **kwargs)


def call_exp(exp: str):
    """
    动态加载包并执行表达式,如:
        a = call_exp('yut.format_date("2024-01-01")')
        o = call_exp('yut.Record(id=100,value=2.0)'

    :param exp: 带完整模块名称的函数及参数
    :return: 函数执行结果
    """
    import re

    def split_args(text: str):
        pattern = r" *(?P<name>.*?)\((?P<args>.*)\)$"
        m = re.search(pattern, text)
        if m:
            return m.group('name'), m.group('args')
        else:
            return None, None

    def split_mod_fn(text: str):
        pattern = r'(?P<mod>.*(?<!\.)\.)?(?P<fn>[^.]+)$'
        m = re.search(pattern, text)
        if m:
            m_mod, m_fn = m.group('mod'), m.group('fn')
            if m_mod:
                m_mod = m_mod[:-1]  # 去掉'.'
            return m_mod, m_fn
        else:
            return None, None

    name, args = split_args(exp)
    if name:
        mod_name, fn_name = split_mod_fn(name)
        if mod_name:
            mod = __import__(mod_name, globals(), locals(), [fn_name])
            _fn_tmp_ = getattr(mod, fn_name)
            return eval(f"_fn_tmp_({args})")
        elif fn_name:  # 不需要import,直接使用default package
            return eval(f"{fn_name}({args})")
    return None


def same_year(d1, d2):
    if not d1 or not d2:
        return ''
    if str(d1)[:4] == str(d2)[:4]:
        return '是'
    return ''


def birth_age(birth, today=None):
    import datetime
    if birth is None:
        return 0
    birth_day = birth
    if isinstance(birth, str):
        birth_day = to_date(birth)
    today_d = today if today else datetime.date.today()
    birth_d = datetime.date(birth_day.year, today_d.month, today_d.day)
    if birth_day <= birth_d:
        age = today_d.year - birth_day.year
    else:
        age = today_d.year - birth_day.year - 1
    return age


def year_date_pair(date):
    """
    返回给定日期对应的年初和年末日期。如： year_date_pair('2024-11-20') -> (2024-01-01,2024-12-31)
    :param date: 日期，可以是str或datetime.date
    :return: (datetime.date,datetime.date)
    """
    y = to_date(date).year
    return to_date('%4d-01-01' % y), to_date('%4d-12-31' % y)


def month_date_pair(y, m):
    """
    返回给定年,月对应的月初和月末日期。如： month_date_pair(2024,11) -> (2024-11-01,2024-11-30)
    :param y: 年份
    :param m: 月度
    :return: (datetime.date,datetime.date)
    """
    import datetime
    start_date = to_date('%4d-%02d-01' % (y, m))
    if m >= 12:
        end_date = to_date('%4d-%02d-01' % (y + 1, 1))
    else:
        end_date = to_date('%4d-%02d-01' % (y, m + 1))
    end_date = end_date - datetime.timedelta(days=1)
    return start_date, end_date


def inc_date(d_str, days=1):
    d = to_date(d_str)
    dt = d + datetime.timedelta(days=days)
    return format_date(dt)


class Obj:
    def __init__(self, **kwargs):
        for k in kwargs:
            v = kwargs[k]
            setattr(self, k, v)

    def __attr_values(self, use_repr=False, skip_prefix='__'):
        ret = ''
        for attr in self.__dict__:
            if not attr.startswith(skip_prefix):
                if use_repr:
                    ret += ' %s=%s;' % (attr, getattr(self, attr).__repr__())
                else:
                    ret += ' %s=%s;' % (attr, getattr(self, attr))
        return ret

    def __getitem__(self, item):
        return self.__dict__.get(item)

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __str__(self):
        return '<%s:%s>' % (self.__class__.__name__, self.__attr_values(use_repr=False))

    def from_obj(self, o):
        if hasattr(o, '__dict__'):
            for attr in o.__dict__:
                if not attr.startswith('__'):
                    setattr(self, attr, getattr(o, attr))
        elif isinstance(o, dict):  # 从字典复制对象
            for k in o.keys():
                v = o.get(k)
                if v is not None:
                    setattr(self, k, v)
                else:
                    setattr(self, k, None)
        return self


def d2o(d, cls=Obj):
    o = cls()
    for k in d.keys():
        v = d.get(k)
        if v is not None:
            setattr(o, k, v)
        else:
            setattr(o, k, None)
    return o


def o2d(o, keys=None):
    """
    将对象属性值转换为字典
    :param o: 待转换的对象
    :param keys: 需要包含的属性，默认为None，表示所有属性
    :return:
    """
    d = {}
    if keys is None:
        keys = o.__dict__.keys()
    for k in keys:
        if k.startswith('*'):  # 使用属性值的str，不是对象
            attr = k[1:]
            d[attr] = str(getattr(o, attr))
        else:
            d[k] = getattr(o, k)
    return d


def r2d(r):
    if hasattr(r, '_asdict'):
        return CaseInsensitiveDict(r._asdict())
    else:
        return CaseInsensitiveDict(r.__dict__)


def json_dump(obj, indent=2, skip_prefix='__', ensure_ascii=False):
    """
    对象或对象集合转换为json串
    :param obj:
    :param indent:
    :param skip_prefix:
    :param ensure_ascii:
    :return:
    """
    from json import JSONEncoder

    class Encoder(JSONEncoder):
        def default(self, o):
            if type(o) in [tuple, set]:
                return list(o)
            if hasattr(o, '__dict__'):
                d = {}
                for key in o.__dict__:
                    if not key.startswith(skip_prefix):
                        value = getattr(o, key)
                        d[key] = value
                return d
            else:
                return str(o)

    return json.dumps(obj, cls=Encoder, indent=indent, ensure_ascii=ensure_ascii)


def json_load(s, cls=Obj):
    """
    从json串中解析对象（或对象集合）
    :param s:
    :param cls:
    :return:
    """
    o = json.loads(s, strict=False)
    o_type = type(o)
    if o_type in (list, tuple, set):
        return [d2o(oo, cls) for oo in o]
    elif o_type == dict:
        return {oo.key: d2o(oo, cls) for oo in o}
    else:
        return d2o(o)


# 大小写无关的字典
class CaseInsensitiveDict(dict):
    """
    大小写无关的字典
    """

    def __getitem__(self, key):
        for k in self.keys():
            if k.lower() == key.lower():
                return dict.__getitem__(self, k)
        raise KeyError(key)

    def __getattr__(self, key):
        for k in self.keys():
            if k.lower() == key.lower():
                return dict.__getitem__(self, k)
        return dict.__getattribute__(self, key)


def change_file_ext(file_name, tar_ext):
    if tar_ext[0] == '.':
        tar_ext = tar_ext[1:]
    fn, ext = os.path.splitext(file_name)
    if ext == tar_ext:
        return f'{fn}_{ext}.{tar_ext}'
    else:
        return f'{fn}.{tar_ext}'


def m64_enc(s: str) -> str:
    bt = base64.b64encode(s.encode('utf-8')).decode('utf-8')
    ret = ''.join([m64_char[b64_char.index(ch)] for ch in bt])
    return base64.b64encode(ret.encode('utf-8')).decode('utf-8')


def m64_dec(s: str, esc='?') -> str:
    if not s:
        return ''
    if s.startswith(esc):  # 明文存储
        return s[len(esc):]
    mt = base64.b64decode(s).decode('utf-8')
    ret = ''.join([b64_char[m64_char.index(ch)] for ch in mt])
    return base64.b64decode(ret).decode('utf-8')


def sec_hms(seconds: int) -> str:
    """
    将秒按照时分秒显示
    :param seconds: 秒
    :return: f'{h:3d}h{m:02d}m{s:02d}s'
    """
    if seconds < 60:
        return f'{seconds:3d}s'
    elif seconds < 3600:
        m = int(seconds / 60)
        s = int(seconds % 60)
        return f'{m:3d}m{s:d}s'
    else:
        h = int(seconds / 3600)
        m = int((seconds % 3600) / 60)
        s = int(seconds % 60)
        return f'{h:3d}h{m:02d}m{s:02d}s'



# -*- py_file.py: python ; coding: utf-8 -*-
P_3_ = ":/pic/3_.png"
P_3DOTS_H = ":/pic/3dots_h.png"
P_3DOTS_V = ":/pic/3dots_v.png"
P_3I = ":/pic/3i.png"
P_4_DOTS = ":/pic/4_dots.png"
P_4_SQUARE = ":/pic/4_square.png"
P_ADD_SUBS = ":/pic/add_subs.png"
P_ADDGREEN = ":/pic/addgreen.png"
P_APPICON = ":/pic/appicon.png"
P_AR_ARROW = ":/pic/ar_arrow.png"
P_AR_CHECK = ":/pic/ar_check.png"
P_ARROW_CURSOR = ":/pic/arrow_cursor.png"
P_ATTACH = ":/pic/attach.png"
P_BACK = ":/pic/back.png"
P_BALL = ":/pic/ball.png"
P_BALL2 = ":/pic/ball2.png"
P_BALL_BLUE = ":/pic/ball_blue.png"
P_BALL_BULLET = ":/pic/ball_bullet.png"
P_BALL_DARKRED = ":/pic/ball_darkred.png"
P_BALL_GOLDEN = ":/pic/ball_golden.png"
P_BALL_GREEN = ":/pic/ball_green.png"
P_BALL_LIGHT = ":/pic/ball_light.png"
P_BALL_RED = ":/pic/ball_red.png"
P_BIZ_1 = ":/pic/biz_1.png"
P_BIZ_2 = ":/pic/biz_2.png"
P_BLUE_STAR = ":/pic/blue_star.png"
P_BULB = ":/pic/bulb.png"
P_BULLET1 = ":/pic/bullet1.png"
P_BULLET1_GOLDEN = ":/pic/bullet1_golden.png"
P_BULLET2 = ":/pic/bullet2.png"
P_BULLET2_GOLDEN = ":/pic/bullet2_golden.png"
P_BULLET3 = ":/pic/bullet3.png"
P_BULLET3_GOLDEN = ":/pic/bullet3_golden.png"
P_BULLET4 = ":/pic/bullet4.png"
P_CALC = ":/pic/calc.png"
P_CALENDAR = ":/pic/calendar.png"
P_CANCEL = ":/pic/cancel.png"
P_CHART = ":/pic/chart.png"
P_CHART_BAR = ":/pic/chart_bar.png"
P_CHART_LINE = ":/pic/chart_line.png"
P_CHART_LINEBAR = ":/pic/chart_linebar.png"
P_CHART_PIE = ":/pic/chart_pie.png"
P_CHAT = ":/pic/chat.png"
P_CHECK = ":/pic/check.png"
P_CHECK_SHEET = ":/pic/check_sheet.png"
P_CIRCLE_CUBE = ":/pic/circle_cube.png"
P_CIRCLE_CUBE2 = ":/pic/circle_cube2.png"
P_CLIP_BOARD = ":/pic/clip_board.png"
P_CLOSE = ":/pic/close.png"
P_CLOSE_X = ":/pic/close_x.png"
P_CONTINUE = ":/pic/continue.png"
P_CONTRACT = ":/pic/contract.png"
P_CONTRACT_2 = ":/pic/contract_2.png"
P_CUBE4 = ":/pic/cube4.png"
P_CUSTOM = ":/pic/custom.png"
P_CUSTOM_CARE = ":/pic/custom_care.png"
P_CUSTOM_MONEY = ":/pic/custom_money.png"
P_DB = ":/pic/db.png"
P_DCUBE = ":/pic/dcube.png"
P_DIAMOND = ":/pic/diamond.png"
P_DOC = ":/pic/doc.png"
P_DOC2 = ":/pic/doc2.png"
P_DOC_PERCENT = ":/pic/doc_percent.png"
P_DOC_RPT = ":/pic/doc_rpt.png"
P_DOWNLOAD = ":/pic/download.png"
P_DOWNLOAD_2 = ":/pic/download_2.png"
P_DROP = ":/pic/drop.png"
P_EARTH = ":/pic/earth.png"
P_EXEC_ALL = ":/pic/exec_all.png"
P_EXECUTE = ":/pic/execute.png"
P_EXPORT = ":/pic/export.png"
P_EYE = ":/pic/eye.png"
P_FANGSHENG = ":/pic/fangsheng.png"
P_FAPICON = ":/pic/fapicon.png"
P_FILTER = ":/pic/filter.png"
P_FIT_WIDTH = ":/pic/fit_width.png"
P_FLY_UP = ":/pic/fly_up.png"
P_FORWARD = ":/pic/forward.png"
P_FUNC = ":/pic/func.png"
P_FUNC_GROUP = ":/pic/func_group.png"
P_FUNC_ITEM = ":/pic/func_item.png"
P_FUNCTION = ":/pic/function.png"
P_GAUGE = ":/pic/gauge.png"
P_GOAL = ":/pic/goal.png"
P_GOLDEN_STAR = ":/pic/golden_star.png"
P_GREEN_STAR = ":/pic/green_star.png"
P_GRID = ":/pic/grid.png"
P_GRID_1 = ":/pic/grid_1.png"
P_GRID_CUBE = ":/pic/grid_cube.png"
P_GRIDVIEW = ":/pic/gridview.png"
P_HAND = ":/pic/hand.png"
P_HEART = ":/pic/heart.png"
P_HEXSTAR = ":/pic/hexstar.png"
P_HOME = ":/pic/home.png"
P_HTML = ":/pic/html.png"
P_INDEX = ":/pic/index.png"
P_INOUT = ":/pic/inout.png"
P_JOIN = ":/pic/join.png"
P_KEY = ":/pic/key.png"
P_LABEL = ":/pic/label.png"
P_LABEL_2 = ":/pic/label_2.png"
P_LAST = ":/pic/last.png"
P_LEFT = ":/pic/left.png"
P_LEFT2 = ":/pic/left2.png"
P_LEVELS = ":/pic/levels.png"
P_LIGHTNING = ":/pic/lightning.png"
P_LINK = ":/pic/link.png"
P_LOCK = ":/pic/lock.png"
P_LOGO = ":/pic/logo.png"
P_LOOKUP = ":/pic/lookup.png"
P_LOOP = ":/pic/loop.png"
P_MAX = ":/pic/max.png"
P_MBLOCKS = ":/pic/mblocks.png"
P_MILESTONE = ":/pic/milestone.png"
P_MIN = ":/pic/min.png"
P_MINUS = ":/pic/minus.png"
P_MONEY_FORWARD = ":/pic/money_forward.png"
P_MORE = ":/pic/more.png"
P_MULTI_BLOCK = ":/pic/multi_block.png"
P_NEXT = ":/pic/next.png"
P_NOTEBOOK = ":/pic/notebook.png"
P_NSTC = ":/pic/nstc.png"
P_OFF = ":/pic/off.png"
P_OKR = ":/pic/okr.png"
P_OPEN = ":/pic/open.png"
P_ORG = ":/pic/org.png"
P_ORG1 = ":/pic/org1.png"
P_ORG2 = ":/pic/org2.png"
P_OUTCOME = ":/pic/outcome.png"
P_PAPER_PLANE = ":/pic/paper_plane.png"
P_PARTS = ":/pic/parts.png"
P_PAUSE = ":/pic/pause.png"
P_PC = ":/pic/pc.png"
P_PC_BAR = ":/pic/pc_bar.png"
P_PERSON_CARD = ":/pic/person_card.png"
P_PERSON_CHART = ":/pic/person_chart.png"
P_PERSON_DOC = ":/pic/person_doc.png"
P_PERSON_FIRE = ":/pic/person_fire.png"
P_PERSON_HIRE = ":/pic/person_hire.png"
P_PIE = ":/pic/pie.png"
P_PIE_2 = ":/pic/pie_2.png"
P_PLAY = ":/pic/play.png"
P_PLUS = ":/pic/plus.png"
P_RED_STAR = ":/pic/red_star.png"
P_REFRESH = ":/pic/refresh.png"
P_REPORT = ":/pic/report.png"
P_REVERT = ":/pic/revert.png"
P_RIGHT = ":/pic/right.png"
P_RIGHT2 = ":/pic/right2.png"
P_RIGHT_3 = ":/pic/right_3.png"
P_SAVE = ":/pic/save.png"
P_SAVE_DATA = ":/pic/save_data.png"
P_SETTING = ":/pic/setting.png"
P_SHOW_NORMAL = ":/pic/show_normal.png"
P_SITE = ":/pic/site.png"
P_SITE_BLUE = ":/pic/site_blue.png"
P_SORT = ":/pic/sort.png"
P_STAR = ":/pic/star.png"
P_STOP = ":/pic/stop.png"
P_SUM = ":/pic/sum.png"
P_SUM_1 = ":/pic/sum_1.png"
P_TARGET = ":/pic/target.png"
P_TIME = ":/pic/time.png"
P_TO_CLOUD = ":/pic/to_cloud.png"
P_TODO = ":/pic/todo.png"
P_TOOLS = ":/pic/tools.png"
P_TREE = ":/pic/tree.png"
P_TRUSH = ":/pic/trush.png"
P_TUMBER = ":/pic/tumber.png"
P_TURN_UP = ":/pic/turn_up.png"
P_TWO_PERSONS = ":/pic/two_persons.png"
P_USER = ":/pic/user.png"
P_VAR = ":/pic/var.png"
P_VIEW = ":/pic/view.png"
P_VV_DOWN = ":/pic/vv_down.png"
P_VVV_DOWN = ":/pic/vvv_down.png"
I_AIRPLANE = ":/icon/airplane.ico"
I_CALC = ":/icon/calc.ico"
I_CIRCLE_CUBE2 = ":/icon/circle_cube2.ico"
I_DCUBE = ":/icon/dcube.ico"
I_FAP = ":/icon/fap.ico"
I_NDAS = ":/icon/ndas.ico"
I_NDW = ":/icon/ndw.ico"

